"""GeckoLogStruct - A class to manage the LogStruct for 'MrSteam v2'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoTempStructAccessor,
    GeckoWordStructAccessor,
)


class GeckoLogStruct:
    """Log Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the log struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the log struct class version."""
        return 2

    @property
    def begin(self) -> int:
        """Get the offset start."""
        return 256

    @property
    def end(self) -> int:
        """Get the offset end."""
        return 301

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return []

    @property
    def all_device_keys(self) -> list[str]:
        """Get all device keys."""
        return []

    @property
    def user_demand_keys(self) -> list[str]:
        """Get all user demand keys."""
        return []

    @property
    def error_keys(self) -> list[str]:
        """Get all error keys."""
        return [
            "FlashErr",
            "H2O2Err",
            "KeyStuckErr",
            "PowerFailErr",
            "Prr1Err",
            "Prr2Err",
            "Prr3Err",
            "Prr4Err",
            "SlaveH2O2Err",
        ]

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "UserMode": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Controls/UserMode",
                256,
                None,
                ["OFF", "ON", "DIAGNOSTIC"],
                None,
                None,
                "ALL",
            ),
            "UserPause": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Controls/UserPause", 257, 1, "ALL"
            ),
            "UserAroma": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Controls/UserAroma",
                258,
                None,
                ["OFF", "ON"],
                None,
                None,
                "ALL",
            ),
            "UserChroma": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Controls/UserChroma",
                259,
                None,
                ["OFF", "ON"],
                None,
                None,
                "ALL",
            ),
            "UserSetpointG": GeckoTempStructAccessor(
                self.struct, "LogStructure/Controls/UserSetpointG", 264, "ALL"
            ),
            "UserRuntime": GeckoWordStructAccessor(
                self.struct, "LogStructure/Controls/UserRuntime", 261, "ALL"
            ),
            "UserProg": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Controls/UserProg",
                263,
                None,
                ["User", "Prog1", "Prog2"],
                None,
                None,
                "ALL",
            ),
            "Hours": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/Hours", 268, None
            ),
            "ModeState": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/ModeState",
                269,
                0,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "PauseState": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/PauseState", 269, 1, None
            ),
            "DiagnosticState": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/DiagnosticState",
                269,
                2,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "ExternalProbe": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/ExternalProbe", 269, 3, None
            ),
            "WaterDetected": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/WaterDetected", 269, 4, None
            ),
            "NoRegulation": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/NoRegulation", 269, 5, None
            ),
            "MasterSlave": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/MasterSlave",
                269,
                6,
                ["SLAVE", "MASTER"],
                None,
                2,
                None,
            ),
            "KeypadProbe": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/KeypadProbe", 270, 0, None
            ),
            "ExpressCycle": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/ExpressCycle", 270, 1, None
            ),
            "SlaveOnState": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/SlaveOnState", 271, 1, None
            ),
            "SlaveHeaterState": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/SlaveHeaterState",
                271,
                3,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "PowerFailErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/PowerFailErr", 272, 0, None
            ),
            "Prr2Err": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Prr2Err", 272, 1, None
            ),
            "Prr1Err": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Prr1Err", 272, 2, None
            ),
            "H2O2Err": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/H2O2Err", 272, 3, None
            ),
            "SlaveH2O2Err": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/SlaveH2O2Err", 272, 4, None
            ),
            "KeyStuckErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/KeyStuckErr", 272, 5, None
            ),
            "FlashErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/FlashErr", 272, 6, None
            ),
            "Prr3Err": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Prr3Err", 272, 7, None
            ),
            "Prr4Err": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Prr4Err", 273, 0, None
            ),
            "Jumper9": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Jumper9", 274, 0, None
            ),
            "Jumper2": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Jumper2", 274, 1, None
            ),
            "Jumper3": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Jumper3", 274, 2, None
            ),
            "Jumper4": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Jumper4", 274, 3, None
            ),
            "Jumper5": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Jumper5", 274, 4, None
            ),
            "Jumper6": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Jumper6", 274, 5, None
            ),
            "Jumper7": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Jumper7", 274, 6, None
            ),
            "Jumper8": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Jumper8", 274, 7, None
            ),
            "MaxRuntime": GeckoWordStructAccessor(
                self.struct, "LogStructure/Status/MaxRuntime", 275, None
            ),
            "KeypadType": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/KeypadType",
                277,
                None,
                ["NO_TSC", "SC_54", "TSC_53", "AUX_SW", "COLOR_SERIES"],
                None,
                None,
                None,
            ),
            "KeypadID": GeckoWordStructAccessor(
                self.struct, "LogStructure/Status/KeypadID", 298, None
            ),
            "KeypadRev": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/KeypadRev", 300, None
            ),
            "KeypadRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/KeypadRel", 301, None
            ),
            "RoomTempG": GeckoTempStructAccessor(
                self.struct, "LogStructure/Status/RoomTempG", 278, None
            ),
            "K1000TempG": GeckoTempStructAccessor(
                self.struct, "LogStructure/Status/K1000TempG", 296, "ALL"
            ),
            "RemainingRuntime": GeckoWordStructAccessor(
                self.struct, "LogStructure/Status/RemainingRuntime", 293, None
            ),
            "DrainValveOutput": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/DrainValveOutput", 295, 0, None
            ),
            "AromaOutput": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/AromaOutput", 295, 1, None
            ),
            "HeaterOutput": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/HeaterOutput", 295, 2, None
            ),
            "WaterValveOutput": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/WaterValveOutput", 295, 3, None
            ),
            "ChromaOutput": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/ChromaOutput", 295, 4, None
            ),
            "PackBootID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/PackBootID", 281, None
            ),
            "PackBootRev": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackBootRev", 283, None
            ),
            "PackBootRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackBootRel", 284, None
            ),
            "PackType": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/PackType",
                285,
                None,
                [
                    "Unknown",
                    "inXE",
                    "MasIBC",
                    "MIA",
                    "DJS4",
                    "inClear",
                    "inXM",
                    "K600",
                    "inTerface",
                    "inTouch",
                    "inYT",
                    "Color_Keypad",
                    "inYJ",
                    "MrSteam",
                ],
                None,
                None,
                None,
            ),
            "PackMemRange": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/PackMemRange",
                286,
                0,
                ["16K", "32K", "48K", "64K"],
                None,
                4,
                None,
            ),
            "PackCoreID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/PackCoreID", 287, None
            ),
            "PackCoreRev": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackCoreRev", 289, None
            ),
            "PackCoreRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackCoreRel", 290, None
            ),
            "PackConfigLib": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackConfigLib", 291, None
            ),
            "PackStatusLib": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackStatusLib", 292, None
            ),
        }



"""Gecko Lights."""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Any

from geckolib.automation.power import GeckoPower
from geckolib.const import GeckoConstants

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade


class GeckoLight(GeckoPower):
    """The base class of lights."""

    def __init__(self, facade: GeckoAsyncFacade, name: str, key: str) -> None:
        """Initialize the light class."""
        super().__init__(facade, name, key)
        self.device_class = GeckoConstants.DEVICE_CLASS_LIGHT

    @property
    @abstractmethod
    def is_on(self) -> bool:
        """Determine if the light is on or not."""
        raise NotImplementedError

    @property
    @abstractmethod
    def state(self) -> str:
        """Get the state of the light."""
        raise NotImplementedError

    async def async_turn_on(self, **kwargs: Any) -> None:
        """Turn the light ON, but does nothing if it is already ON."""
        raise NotImplementedError

    async def async_turn_off(self) -> None:
        """Turn the device OFF, but does nothing if it is already OFF."""
        raise NotImplementedError

    def __str__(self) -> str:
        """Stringize."""
        return f"{self.name}: {self.state}"

    @property
    def monitor(self) -> str:
        """Get a monitor string."""
        return f"{self.key}: {self.state}"


class GeckoLightLi(GeckoLight):
    """Class for the Li light."""

    def __init__(self, facade: GeckoAsyncFacade) -> None:
        """Initialize the Li light."""
        super().__init__(facade, "Light", "LI")
        if "LI" in self.facade.spa.struct.connections:
            self.set_availability(is_available=True)

        if self.is_available:
            self._state_accessor = self.facade.spa.accessors["UdLi"]
            self._state_accessor.watch(self._on_change)

    @property
    def is_on(self) -> bool:
        """Determine if the light is on or not."""
        return self.state != "OFF"

    @property
    def state(self) -> str:
        """Get the state of the light."""
        return self._state_accessor.value

    async def async_turn_on(self, **_kwargs: Any) -> None:
        """Turn the light ON, but does nothing if it is already ON."""
        if self.is_on:
            return
        await self._state_accessor.async_set_value("HI")

    async def async_turn_off(self, **_kwargs: Any) -> None:
        """Turn the light OFF, but does nothing if it is already OFF."""
        if not self.is_on:
            return
        await self._state_accessor.async_set_value("OFF")


class GeckoLightL120(GeckoLight):
    """Class for the L120 light."""

    def __init__(self, facade: GeckoAsyncFacade) -> None:
        """Initialize the L120 light."""
        super().__init__(facade, "Light 2", "L120")
        if "L120" in self.facade.spa.struct.connections:
            self.set_availability(is_available=True)

        if self.is_available:
            self._state_accessor = self.facade.spa.accessors["UdL120"]
            self._state_accessor.watch(self._on_change)

    @property
    def is_on(self) -> bool:
        """Determine if the light is on or not."""
        return self.state != "OFF"

    @property
    def state(self) -> str:
        """Get the state of the light."""
        return self._state_accessor.value

    async def async_turn_on(self, **_kwargs: Any) -> None:
        """Turn the light ON, but does nothing if it is already ON."""
        if self.is_on:
            return
        await self._state_accessor.async_set_value("ON")

    async def async_turn_off(self, **_kwargs: Any) -> None:
        """Turn the light OFF, but does nothing if it is already OFF."""
        if not self.is_on:
            return
        await self._state_accessor.async_set_value("OFF")



"""Automation inGrid class."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from geckolib.const import GeckoConstants

from .select import GeckoSelect

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade
    from geckolib.driver.accessor import GeckoBoolStructAccessor

_LOGGER = logging.getLogger(__name__)


class GeckoInGrid(GeckoSelect):
    """The inGrid mode."""

    def __init__(self, facade: GeckoAsyncFacade) -> None:
        """Initialize the inGrid class."""
        super().__init__(facade, "Heating Management", GeckoConstants.KEY_COOLZONE_MODE)
        if GeckoConstants.KEY_INGRID_DETECTED not in self._spa.accessors:
            self.set_availability(is_available=False)
        else:
            in_grid_detected: GeckoBoolStructAccessor = self._spa.accessors[
                GeckoConstants.KEY_INGRID_DETECTED
            ]
            if not in_grid_detected.value:
                self.set_availability(is_available=False)
        # Set of mappings of constants to UI options
        self.set_mapping(
            {
                "INTERNAL_HEAT": "Electrical heater",
                "HEAT_SAVER": "External heater",
                "BOTH_HEAT": "Both",
                "HEAT_W_BOOST": "Smart",
            }
        )

    @property
    def state(self) -> str:
        """Get the current state via the mapping."""
        if self._accessor is None:
            return "(Unknown)"
        return self.mapping[self._accessor.value]

    async def async_set_state(self, new_state: str) -> None:
        """Set the state of the select entity."""
        if self._accessor is None:
            _LOGGER.warning("%s can't set state with no accessor", self)
            return
        if new_state in self.reverse:
            new_state = self.reverse[new_state]
        await self._accessor.async_set_value(new_state)

    @property
    def states(self) -> list[str]:
        """Get the possible states."""
        if self._accessor is None:
            return []
        return list(self.mapping.values())



"""Gecko Reminders."""

from __future__ import annotations

import logging
from datetime import UTC, datetime
from typing import TYPE_CHECKING

from geckolib.driver import GeckoReminderType

from .base import GeckoAutomationFacadeBase

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade

logger = logging.getLogger(__name__)


class GeckoReminders(GeckoAutomationFacadeBase):
    """Reminders management."""

    class Reminder:
        """A single reminder instance."""

        def __init__(self, rem: tuple[GeckoReminderType, int]) -> None:
            """Initialize a single reminder."""
            self._reminder_type: GeckoReminderType = rem[0]
            self._days: int = rem[1]

        @property
        def reminder_type(self) -> GeckoReminderType:
            """Get the reminder type."""
            return self._reminder_type

        @property
        def description(self) -> str:
            """Get the reminder description."""
            return GeckoReminderType.to_string(self.reminder_type)

        @property
        def days(self) -> int:
            """Get the remaining days."""
            return self._days

        def set_days(self, days: int) -> None:
            """Set the remaining days."""
            self._days = days

        @property
        def monitor(self) -> str:
            """Get the monitor string."""
            return f"{datetime.now(tz=UTC)}"

        def __str__(self) -> str:
            """Stringize the reminder."""
            return (
                f"{self.description} due in {self.days} days"
                if self.days > 0
                else f"{self.description} due today"
                if self.days == 0
                else f"{self.description} overdue by {-self.days} days"
            )

    def __init__(self, facade: GeckoAsyncFacade) -> None:
        """Initialize the reminders class."""
        super().__init__(facade, "Reminders", "REMINDERS")

        self._all_reminders: list[GeckoReminders.Reminder] = []
        self._reminders_handler = None
        self._last_update = None
        if facade.spa.struct.is_spa_pack:
            self.set_availability(is_available=True)

    @property
    def reminders(self) -> list[Reminder]:
        """Return active reminders."""
        return [
            reminder
            for reminder in self._all_reminders
            if reminder.reminder_type != GeckoReminderType.INVALID
        ]

    def get_reminder(
        self, reminder_type: GeckoReminderType
    ) -> GeckoReminders.Reminder | None:
        """Get the reminder of the specified type, or None if not found."""
        for reminder in self._all_reminders:
            if reminder.reminder_type == reminder_type:
                return reminder
        return None

    async def set_reminder(self, reminder_type: GeckoReminderType, days: int) -> None:
        """Set the remaining days for the specified reminder type."""
        for reminder in self._all_reminders:
            if reminder.reminder_type == reminder_type:
                reminder.set_days(days)
        await self.facade.spa.async_set_reminders(
            [
                (reminder.reminder_type, reminder.days)
                for reminder in self._all_reminders
            ]
        )
        self._on_change(self)

    @property
    def last_update(self) -> datetime | None:
        """Time of last reminder update."""
        return self._last_update

    def change_reminders(self, reminders: list[tuple]) -> None:
        """Call from async facade to update active reminders."""
        self._last_update = datetime.now(tz=UTC)
        self._all_reminders = [
            GeckoReminders.Reminder(reminder) for reminder in reminders
        ]
        self._on_change(self)

    def __str__(self) -> str:
        """Stringize the class."""
        if self.reminders is None:
            return f"{self.name}: Waiting..."
        return f"{self.name}: {self.reminders}"



"""Library to communicate with Gecko Alliance products via the in.touch2 module."""

import logging

from ._version import VERSION
from .async_locator import GeckoAsyncLocator
from .async_spa import GeckoAsyncSpa
from .async_spa_descriptor import GeckoAsyncSpaDescriptor
from .async_spa_manager import GeckoAsyncSpaMan
from .async_taskman import GeckoAsyncTaskMan
from .async_tasks import AsyncTasks
from .automation import (
    BainUltra,
    GeckoAsyncFacade,
    GeckoAutomationBase,
    GeckoAutomationFacadeBase,
    GeckoBinarySensor,
    GeckoBlower,
    GeckoBubbleGenerator,
    GeckoButton,
    GeckoErrorSensor,
    GeckoHeatPump,
    GeckoInGrid,
    GeckoInMix,
    GeckoInMixSynchro,
    GeckoInMixZone,
    GeckoKeypad,
    GeckoLight,
    GeckoNumber,
    GeckoPump,
    GeckoReminders,
    GeckoSensor,
    GeckoSwitch,
    GeckoWaterCare,
    GeckoWaterfall,
    GeckoWaterHeater,
    GeckoWaterHeaterAbstract,
    MrSteam,
)
from .config import GeckoConfig
from .const import GeckoConstants
from .driver import (
    GeckoAsyncPartialStatusBlockProtocolHandler,
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoConfigFileProtocolHandler,
    GeckoEnumStructAccessor,
    GeckoGetChannelProtocolHandler,
    GeckoGetWatercareModeProtocolHandler,
    GeckoGetWatercareScheduleListProtocolHandler,
    GeckoHelloProtocolHandler,
    GeckoPackCommandProtocolHandler,
    GeckoPacketProtocolHandler,
    GeckoPingProtocolHandler,
    GeckoRemindersProtocolHandler,
    GeckoReminderType,
    GeckoSetWatercareModeProtocolHandler,
    GeckoStatusBlockProtocolHandler,
    GeckoStructAccessor,
    GeckoUdpProtocolHandler,
    GeckoUpdateFirmwareProtocolHandler,
    GeckoVersionProtocolHandler,
    GeckoWatercareScheduleManager,
    GeckoWordStructAccessor,
    Observable,
)
from .spa_events import GeckoSpaEvent
from .spa_state import GeckoSpaState
from .utils import CUI, GeckoShell, GeckoSimulator, GeckoSnapshot

# Module logger, uses the library name (at this time it was geckolib) and it
# is silent unless required ...
logging.getLogger(__name__).addHandler(logging.NullHandler())

__version__ = VERSION

__all__ = [
    "CUI",
    "VERSION",
    "AsyncTasks",
    "BainUltra",
    "GeckoAsyncFacade",
    "GeckoAsyncLocator",
    "GeckoAsyncPartialStatusBlockProtocolHandler",
    "GeckoAsyncSpa",
    "GeckoAsyncSpaDescriptor",
    "GeckoAsyncSpaMan",
    "GeckoAsyncStructure",
    "GeckoAsyncTaskMan",
    "GeckoAutomationBase",
    "GeckoAutomationFacadeBase",
    "GeckoBinarySensor",
    "GeckoBlower",
    "GeckoBoolStructAccessor",
    "GeckoBubbleGenerator",
    "GeckoButton",
    "GeckoByteStructAccessor",
    "GeckoConfig",
    "GeckoConfigFileProtocolHandler",
    "GeckoConstants",
    "GeckoEnumStructAccessor",
    "GeckoErrorSensor",
    "GeckoGetChannelProtocolHandler",
    "GeckoGetWatercareModeProtocolHandler",
    "GeckoGetWatercareScheduleListProtocolHandler",
    "GeckoHeatPump",
    "GeckoHelloProtocolHandler",
    "GeckoInGrid",
    "GeckoInMix",
    "GeckoInMixSynchro",
    "GeckoInMixZone",
    "GeckoKeypad",
    "GeckoLight",
    "GeckoNumber",
    "GeckoPackCommandProtocolHandler",
    "GeckoPacketProtocolHandler",
    "GeckoPingProtocolHandler",
    "GeckoPump",
    "GeckoReminderType",
    "GeckoReminders",
    "GeckoRemindersProtocolHandler",
    "GeckoSensor",
    "GeckoSetWatercareModeProtocolHandler",
    "GeckoShell",
    "GeckoSimulator",
    "GeckoSnapshot",
    "GeckoSpaEvent",
    "GeckoSpaState",
    "GeckoStatusBlockProtocolHandler",
    "GeckoStructAccessor",
    "GeckoSwitch",
    "GeckoUdpProtocolHandler",
    "GeckoUpdateFirmwareProtocolHandler",
    "GeckoVersionProtocolHandler",
    "GeckoWaterCare",
    "GeckoWaterHeater",
    "GeckoWaterHeaterAbstract",
    "GeckoWatercareScheduleManager",
    "GeckoWaterfall",
    "GeckoWordStructAccessor",
    "MrSteam",
    "Observable",
]



"""Automation switches."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from geckolib.automation.power import GeckoPower
from geckolib.const import GeckoConstants

from .base import GeckoAutomationFacadeBase
from .sensors import GeckoSensor

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade
    from geckolib.driver.accessor import GeckoEnumStructAccessor

_LOGGER = logging.getLogger(__name__)


class GeckoSwitch(GeckoPower):
    """A switch can turn something on or off, and can report the current state."""

    def __init__(self, facade: GeckoAsyncFacade, key: str, props: tuple) -> None:
        """Props is a tuple of (name, keypad_button, state_key, device_class)."""
        super().__init__(facade, props[0], key)
        if key in self._spa.accessors:
            self._accessor = self._spa.accessors[key]
        else:
            self._accessor = self._spa.accessors[props[1]]
        self._state_sensor = GeckoSensor(facade, f"{props[0]} State", self._accessor)
        self._state_sensor.watch(self._on_change)
        self.device_class = props[2]

    @property
    def is_on(self) -> bool:
        """True if the device is ON, False otherwise."""
        if self._accessor.accessor_type == GeckoConstants.SPA_PACK_STRUCT_BOOL_TYPE:
            return self._state_sensor.state
        return self._state_sensor.state != "OFF"

    async def async_turn_on(self) -> None:
        """Turn the device ON, but does nothing if it is already ON."""
        _LOGGER.debug("%s async turn ON", self.name)
        if self.is_on:
            _LOGGER.debug("%s request to turn ON ignored, it's already on!", self.name)
            return
        _LOGGER.debug("Set async state on accessor")
        if self._accessor.accessor_type == GeckoConstants.SPA_PACK_STRUCT_BOOL_TYPE:
            await self._accessor.async_set_value(True)  # noqa: FBT003
        else:
            await self._accessor.async_set_value("ON")

    async def async_turn_off(self) -> None:
        """Turn the device OFF, but does nothing if it is already OFF."""
        _LOGGER.debug("%s async turn OFF", self.name)
        if not self.is_on:
            _LOGGER.debug(
                "%s request to turn OFF ignored, it's already off!", self.name
            )
            return
        _LOGGER.debug("Set async state on accessor")
        if self._accessor.accessor_type == GeckoConstants.SPA_PACK_STRUCT_BOOL_TYPE:
            await self._accessor.async_set_value(False)  # noqa: FBT003
        else:
            await self._accessor.async_set_value("OFF")

    def state_sensor(self) -> GeckoSensor:
        """Get the state sensor."""
        return self._state_sensor

    @property
    def state(self) -> str:
        """Get the switch state."""
        if self._accessor.accessor_type == GeckoConstants.SPA_PACK_STRUCT_BOOL_TYPE:
            return "ON" if self._state_sensor.state else "OFF"
        return self._state_sensor.state

    def __str__(self) -> str:
        """Stringize."""
        return f"{self.name}: {self._state_sensor.state}"

    @property
    def monitor(self) -> str:
        """Get a monitor string."""
        return f"{self.key}: {self._state_sensor.state}"


class GeckoStandby(GeckoAutomationFacadeBase):
    """Standby switch."""

    def __init__(self, facade: GeckoAsyncFacade) -> None:
        """Initialize the standby switch."""
        super().__init__(facade, "Standby", GeckoConstants.KEY_STANDBY)
        self._accessor: GeckoEnumStructAccessor = self._spa.accessors[
            GeckoConstants.KEY_STANDBY
        ]
        self._accessor.watch(self._on_change)

    @property
    def is_on(self) -> bool:
        """True if the device is ON, False otherwise."""
        return self._accessor.value == "OFF"

    async def async_turn_on(self) -> None:
        """Turn on standby mode."""
        await self._accessor.async_set_value("OFF")

    async def async_turn_off(self) -> None:
        """Turn off standby mode."""
        await self._accessor.async_set_value("NOT_SET")

    @property
    def state(self) -> str:
        """Get the standby state."""
        return f"{self.is_on}"

    def __str__(self) -> str:
        """Stringize the class."""
        return f"{self.name}: {self.is_on}"

    @property
    def monitor(self) -> str:
        """Get the monitor string."""
        return f"{self.key}: {self.is_on}"



"""Gecko Keypads."""

from __future__ import annotations

from typing import TYPE_CHECKING

from geckolib.automation.button import GeckoButton
from geckolib.automation.keypad_backlight import GeckoKeypadBacklight
from geckolib.const import GeckoConstants

from .base import GeckoAutomationFacadeBase

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade


class GeckoKeypad(GeckoAutomationFacadeBase):
    """Keypad management class."""

    def __init__(self, facade: GeckoAsyncFacade) -> None:
        """Initialize the keypad class."""
        super().__init__(facade, "Keypad", "KEYPAD")

        self.backlight = GeckoKeypadBacklight(facade)

        self._buttons = []

        # Should we show eco mode button too?

        if self.facade.pump_1.is_available:
            self._buttons.append(
                GeckoButton(facade, "Key Pump 1", GeckoConstants.KEYPAD_PUMP_1)
            )
        if self.facade.pump_2.is_available:
            self._buttons.append(
                GeckoButton(facade, "Key Pump 2", GeckoConstants.KEYPAD_PUMP_2)
            )
        if self.facade.pump_3.is_available:
            self._buttons.append(
                GeckoButton(facade, "Key Pump 3", GeckoConstants.KEYPAD_PUMP_3)
            )
        if self.facade.pump_4.is_available:
            self._buttons.append(
                GeckoButton(facade, "Key Pump 4", GeckoConstants.KEYPAD_PUMP_4)
            )
        if self.facade.pump_5.is_available:
            self._buttons.append(
                GeckoButton(facade, "Key Pump 5", GeckoConstants.KEYPAD_PUMP_5)
            )
        if self.facade.blower.is_available:
            self._buttons.append(
                GeckoButton(facade, "Key Blower", GeckoConstants.KEYPAD_BLOWER)
            )
        if self.facade.waterfall.is_available:
            self._buttons.append(
                GeckoButton(facade, "Key Waterfall", GeckoConstants.KEYPAD_WATERFALL)
            )
        if self.facade.bubblegenerator.is_available:
            self._buttons.append(
                GeckoButton(facade, "Key Bubble Generator", GeckoConstants.KEYPAD_AUX)
            )

        if self.facade.light.is_available:
            self._buttons.append(
                GeckoButton(facade, "Key Light", GeckoConstants.KEYPAD_LIGHT)
            )
        if self.facade.light2.is_available:
            self._buttons.append(
                GeckoButton(facade, "Key Light 2", GeckoConstants.KEYPAD_LIGHT_120)
            )

        if self.facade.water_heater.is_available:
            self._buttons.append(
                GeckoButton(facade, "Key Temperature Up", GeckoConstants.KEYPAD_UP)
            )
            self._buttons.append(
                GeckoButton(facade, "Key Temperature Down", GeckoConstants.KEYPAD_DOWN)
            )

    @property
    def buttons(self) -> list[GeckoButton]:
        """Get the buttons that have any action."""
        return self._buttons

    def __str__(self) -> str:
        """Stringize the class."""
        return f"{self.name}:"



"""Configuration management for geckolib."""

import asyncio
import logging
from dataclasses import dataclass

_LOGGER = logging.getLogger(__name__)


@dataclass
class _GeckoConfig:
    DISCOVERY_INITIAL_TIMEOUT_IN_SECONDS = 1
    """Mininum time in seconds to wait for initial spa discovery even
    if one spa has responded"""

    DISCOVERY_TIMEOUT_IN_SECONDS = 1
    """Maximum time in seconds to wait for full discovery if no spas
    have responded"""

    TASK_TIDY_FREQUENCY_IN_SECONDS = 1
    """Time in seconds between task tidyup checks"""

    PING_FREQUENCY_IN_SECONDS = 1
    """Frequency in seconds to ping the spa to ensure it is still available"""

    PING_DEVICE_NOT_RESPONDING_TIMEOUT_IN_SECONDS = 1
    """Time after which a spa is deemed to be not responding to pings"""

    FACADE_UPDATE_FREQUENCY_IN_SECONDS = 1
    """Frequency in seconds to update facade data"""

    SPA_PACK_REFRESH_FREQUENCY_IN_SECONDS = 1
    """Frequency in seconds to request all LOG data from spa"""

    PROTOCOL_TIMEOUT_IN_SECONDS = 1
    """Default timeout for most protocol commands"""

    PAUSE_BETWEEN_RETRIES_IN_SECONDS = 1
    """Default pause between retry operations"""


CONFIG_MEMBERS = [
    attr
    for attr in dir(_GeckoConfig)
    if not callable(getattr(_GeckoConfig, attr)) and not attr.startswith("__")
]


@dataclass
class _GeckoActiveConfig(_GeckoConfig):
    """Gecko active configuration."""

    DISCOVERY_INITIAL_TIMEOUT_IN_SECONDS = 4
    DISCOVERY_TIMEOUT_IN_SECONDS = 10
    TASK_TIDY_FREQUENCY_IN_SECONDS = 5
    PING_FREQUENCY_IN_SECONDS = 2
    PING_DEVICE_NOT_RESPONDING_TIMEOUT_IN_SECONDS = 10
    FACADE_UPDATE_FREQUENCY_IN_SECONDS = 28800
    SPA_PACK_REFRESH_FREQUENCY_IN_SECONDS = 28800
    PROTOCOL_TIMEOUT_IN_SECONDS = 4
    PAUSE_BETWEEN_RETRIES_IN_SECONDS = 0.4


@dataclass
class _GeckoIdleConfig(_GeckoConfig):
    """Gecko idle configuration."""

    DISCOVERY_INITIAL_TIMEOUT_IN_SECONDS = 4
    DISCOVERY_TIMEOUT_IN_SECONDS = 10
    TASK_TIDY_FREQUENCY_IN_SECONDS = 60
    PING_FREQUENCY_IN_SECONDS = 2
    PING_DEVICE_NOT_RESPONDING_TIMEOUT_IN_SECONDS = 120
    FACADE_UPDATE_FREQUENCY_IN_SECONDS = 3600
    SPA_PACK_REFRESH_FREQUENCY_IN_SECONDS = 3600
    PROTOCOL_TIMEOUT_IN_SECONDS = 4
    PAUSE_BETWEEN_RETRIES_IN_SECONDS = 0.4


# Root config
GeckoConfig: _GeckoConfig = _GeckoIdleConfig()
ConfigChangeEvent: asyncio.Event = asyncio.Event()


def release_config_change_waiters() -> None:
    """Release config change waiters."""
    ConfigChangeEvent.set()
    ConfigChangeEvent.clear()


def set_config_mode(*, active: bool) -> None:
    """Set config mode to active (true) or idle (false)."""
    _LOGGER.debug("set_config_mode: %s", active)
    new_config = _GeckoActiveConfig() if active else _GeckoIdleConfig()
    for member in CONFIG_MEMBERS:
        setattr(GeckoConfig, member, getattr(new_config, member))
    release_config_change_waiters()


async def config_sleep(delay: float | None, _reason: str) -> bool:
    """Sleep wrapper that also handles config changes, returns True on timeout."""
    if delay is None:
        await asyncio.sleep(0)
        return False
    try:
        async with asyncio.timeout(delay):
            await ConfigChangeEvent.wait()
    except TimeoutError:
        return True
    return False



"""Automation selection class."""  # noqa: A005

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from .base import GeckoAutomationFacadeBase

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade

_LOGGER = logging.getLogger(__name__)


class GeckoSelect(GeckoAutomationFacadeBase):
    """A select object can select between options and can report the current state."""

    def __init__(self, facade: GeckoAsyncFacade, name: str, tag: str) -> None:
        """Initialize the select class."""
        super().__init__(facade, name, tag)
        self._accessor = None
        if tag in facade.spa.accessors:
            self._accessor = facade.spa.accessors[tag]
            self._accessor.watch(self._on_change)
            self.set_availability(is_available=True)

    @property
    def state(self) -> str:
        """Get the current state."""
        if self._accessor is None:
            return "Unknown"
        return self._accessor.value

    async def async_set_state(self, new_state: str) -> None:
        """Set the state of the select entity."""
        if self._accessor is None:
            _LOGGER.warning("%s can't set state with no accessor.", self)
            return
        _LOGGER.debug("%s async set state %s", self.name, new_state)
        await self._accessor.async_set_value(new_state)

    @property
    def states(self) -> list[str]:
        """Get the possible states."""
        if self._accessor is None:
            return []
        return self._accessor.items

    def __str__(self) -> str:
        """Stringize the class."""
        return f"{self.name}: {self.state}"

    @property
    def monitor(self) -> str:
        """Get monitore string."""
        return f"{self.key}: {self.state}"



"""AsyncCmd class."""

import asyncio
import cmd
import logging
import threading
from queue import Empty, Queue
from typing import Any, Self

_LOGGER = logging.getLogger(__name__)


class AsyncCmd(cmd.Cmd):
    """
    AsyncCmd. Async command processor based on cmd.Cmd.

    This class will run an async loop in a thread and all the async
    functions will run in there, including any do_ commands in the
    derived shell that are decorated with the async keyword.
    """

    prompt = "(async-cmd) "

    def __init__(self) -> None:
        """Initialize the async class."""
        super().__init__()
        self.queue: Queue[tuple] = Queue()
        self.stop_event: threading.Event = threading.Event()
        self.thread: threading.Thread = threading.Thread(
            target=self.run_async_loop, daemon=True
        )
        self.thread.start()

    def __enter__(self) -> Self:
        """Suport python 'with' syntax."""
        _LOGGER.debug("Start shell %s", self.__class__.__name__)
        return self

    def __exit__(self, *exc_info: object) -> None:
        """Suport puthon 'with' syntax."""
        _LOGGER.debug("Stop shell %s", self.__class__.__name__)

    def run_async_loop(self) -> None:
        """Run the async loop."""
        _LOGGER.debug("Running async loop")
        with asyncio.Runner() as runner:
            runner.run(self.process_commands())

    async def process_commands(self) -> None:
        """Process command loop."""
        aenter = self.__getattribute__("__aenter__")
        if aenter is not None:
            _LOGGER.debug("Calling aenter %s", aenter)
            await aenter()

        try:
            while not self.stop_event.is_set():
                try:
                    coroutine, args = self.queue.get(timeout=0.01)
                    _LOGGER.debug("Run coroutine %s with %s", coroutine, args)
                    await coroutine(*args)
                    self.queue.task_done()
                except Empty:
                    await asyncio.sleep(0)
                except asyncio.CancelledError:
                    _LOGGER.debug("Command loop cancelled")
                    raise
                except Exception:
                    _LOGGER.exception("Command loop exception, logged and continue")
        finally:
            aexit = self.__getattribute__("__aexit__")
            if aexit is not None:
                _LOGGER.debug("Calling aexit %s", aexit)
                await aexit()

    def dispatch(self, coroutine: Any, *args: Any) -> None:
        """Dispatch a command the async loop."""
        self.queue.put_nowait((coroutine, args))

    def precmd(self, line: str) -> str:
        """Determine if the command is running in the io loop, or async loop."""
        parts = line.split(" ", 1)
        command_name: str = parts[0]
        rest: str = parts[1] if len(parts) > 1 else ""
        method: Any | None = getattr(self, f"do_{command_name}", None)
        if asyncio.iscoroutinefunction(method):
            self.dispatch(method, rest)
            return ""
        return line

    def do_exit(self, _line: str) -> bool:
        """Stop the async loop and exits the program."""
        self.stop_event.set()
        self.thread.join()
        return True



"""GeckoLogStruct - A class to manage the LogStruct for 'InMix v2'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoWordStructAccessor,
)


class GeckoLogStruct:
    """Log Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the log struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the log struct class version."""
        return 2

    @property
    def begin(self) -> int:
        """Get the offset start."""
        return 600

    @property
    def end(self) -> int:
        """Get the offset end."""
        return 639

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return []

    @property
    def all_device_keys(self) -> list[str]:
        """Get all device keys."""
        return []

    @property
    def user_demand_keys(self) -> list[str]:
        """Get all user demand keys."""
        return []

    @property
    def error_keys(self) -> list[str]:
        """Get all error keys."""
        return []

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "InMix-Synchro1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-Synchro1",
                600,
                None,
                ["ZONE1", "ZONE2", "ZONE3", "ZONE4"],
                None,
                None,
                "ALL",
            ),
            "InMix-Mode1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-Mode1",
                601,
                None,
                [
                    "STATIC",
                    "VERY_SLOW_FADE",
                    "SLOW_FADE",
                    "BENOIT1",
                    "BENOIT2",
                    "RGB",
                    "DELAY_ADJUSTABLE_FADE_MODE",
                    "LIMITED_COLOR_FADE_MODE",
                ],
                None,
                None,
                "ALL",
            ),
            "InMix-Color1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-Color1",
                602,
                None,
                [
                    "NO_COLOR",
                    "WHITE",
                    "RED",
                    "GREEN",
                    "BLUE",
                    "PURPLE",
                    "YELLOW",
                    "CYAN",
                    "ORANGE",
                    "MAGENTA",
                    "ROYAL_BLUE",
                ],
                None,
                None,
                "ALL",
            ),
            "InMix-Speed1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-Speed1",
                603,
                None,
                [
                    "PRESET",
                    "LOW",
                    "MEDIUM",
                    "HIGH",
                    "PAUSE",
                    "FADE_ADJUST_DELAY0",
                    "FADE_ADJUST_DELAY1",
                    "FADE_ADJUST_DELAY2",
                    "FADE_ADJUST_DELAY3",
                    "FADE_ADJUST_DELAY4",
                    "FADE_ADJUST_DELAY5",
                ],
                None,
                None,
                "ALL",
            ),
            "InMix-RedLevel1": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMix-RedLevel1", 604, "ALL"
            ),
            "InMix-GreenLevel1": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMix-GreenLevel1", 605, "ALL"
            ),
            "InMix-BlueLevel1": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMix-BlueLevel1", 606, "ALL"
            ),
            "InMix-Synchro2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-Synchro2",
                607,
                None,
                ["ZONE1", "ZONE2", "ZONE3", "ZONE4"],
                None,
                None,
                "ALL",
            ),
            "InMix-Mode2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-Mode2",
                608,
                None,
                [
                    "STATIC",
                    "VERY_SLOW_FADE",
                    "SLOW_FADE",
                    "BENOIT1",
                    "BENOIT2",
                    "RGB",
                    "DELAY_ADJUSTABLE_FADE_MODE",
                    "LIMITED_COLOR_FADE_MODE",
                ],
                None,
                None,
                "ALL",
            ),
            "InMix-Color2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-Color2",
                609,
                None,
                [
                    "NO_COLOR",
                    "WHITE",
                    "RED",
                    "GREEN",
                    "BLUE",
                    "PURPLE",
                    "YELLOW",
                    "CYAN",
                    "ORANGE",
                    "MAGENTA",
                    "ROYAL_BLUE",
                ],
                None,
                None,
                "ALL",
            ),
            "InMix-Speed2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-Speed2",
                610,
                None,
                [
                    "PRESET",
                    "LOW",
                    "MEDIUM",
                    "HIGH",
                    "PAUSE",
                    "FADE_ADJUST_DELAY0",
                    "FADE_ADJUST_DELAY1",
                    "FADE_ADJUST_DELAY2",
                    "FADE_ADJUST_DELAY3",
                    "FADE_ADJUST_DELAY4",
                    "FADE_ADJUST_DELAY5",
                ],
                None,
                None,
                "ALL",
            ),
            "InMix-RedLevel2": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMix-RedLevel2", 611, "ALL"
            ),
            "InMix-GreenLevel2": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMix-GreenLevel2", 612, "ALL"
            ),
            "InMix-BlueLevel2": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMix-BlueLevel2", 613, "ALL"
            ),
            "InMix-Synchro3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-Synchro3",
                614,
                None,
                ["ZONE1", "ZONE2", "ZONE3", "ZONE4"],
                None,
                None,
                "ALL",
            ),
            "InMix-Mode3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-Mode3",
                615,
                None,
                [
                    "STATIC",
                    "VERY_SLOW_FADE",
                    "SLOW_FADE",
                    "BENOIT1",
                    "BENOIT2",
                    "RGB",
                    "DELAY_ADJUSTABLE_FADE_MODE",
                    "LIMITED_COLOR_FADE_MODE",
                ],
                None,
                None,
                "ALL",
            ),
            "InMix-Color3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-Color3",
                616,
                None,
                [
                    "NO_COLOR",
                    "WHITE",
                    "RED",
                    "GREEN",
                    "BLUE",
                    "PURPLE",
                    "YELLOW",
                    "CYAN",
                    "ORANGE",
                    "MAGENTA",
                    "ROYAL_BLUE",
                ],
                None,
                None,
                "ALL",
            ),
            "InMix-Speed3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-Speed3",
                617,
                None,
                [
                    "PRESET",
                    "LOW",
                    "MEDIUM",
                    "HIGH",
                    "PAUSE",
                    "FADE_ADJUST_DELAY0",
                    "FADE_ADJUST_DELAY1",
                    "FADE_ADJUST_DELAY2",
                    "FADE_ADJUST_DELAY3",
                    "FADE_ADJUST_DELAY4",
                    "FADE_ADJUST_DELAY5",
                ],
                None,
                None,
                "ALL",
            ),
            "InMix-RedLevel3": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMix-RedLevel3", 618, "ALL"
            ),
            "InMix-GreenLevel3": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMix-GreenLevel3", 619, "ALL"
            ),
            "InMix-BlueLevel3": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMix-BlueLevel3", 620, "ALL"
            ),
            "InMix-Synchro4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-Synchro4",
                621,
                None,
                ["ZONE1", "ZONE2", "ZONE3", "ZONE4"],
                None,
                None,
                "ALL",
            ),
            "InMix-Mode4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-Mode4",
                622,
                None,
                [
                    "STATIC",
                    "VERY_SLOW_FADE",
                    "SLOW_FADE",
                    "BENOIT1",
                    "BENOIT2",
                    "RGB",
                    "DELAY_ADJUSTABLE_FADE_MODE",
                    "LIMITED_COLOR_FADE_MODE",
                ],
                None,
                None,
                "ALL",
            ),
            "InMix-Color4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-Color4",
                623,
                None,
                [
                    "NO_COLOR",
                    "WHITE",
                    "RED",
                    "GREEN",
                    "BLUE",
                    "PURPLE",
                    "YELLOW",
                    "CYAN",
                    "ORANGE",
                    "MAGENTA",
                    "ROYAL_BLUE",
                ],
                None,
                None,
                "ALL",
            ),
            "InMix-Speed4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-Speed4",
                624,
                None,
                [
                    "PRESET",
                    "LOW",
                    "MEDIUM",
                    "HIGH",
                    "PAUSE",
                    "FADE_ADJUST_DELAY0",
                    "FADE_ADJUST_DELAY1",
                    "FADE_ADJUST_DELAY2",
                    "FADE_ADJUST_DELAY3",
                    "FADE_ADJUST_DELAY4",
                    "FADE_ADJUST_DELAY5",
                ],
                None,
                None,
                "ALL",
            ),
            "InMix-RedLevel4": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMix-RedLevel4", 625, "ALL"
            ),
            "InMix-GreenLevel4": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMix-GreenLevel4", 626, "ALL"
            ),
            "InMix-BlueLevel4": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMix-BlueLevel4", 627, "ALL"
            ),
            "InMix-FadeIntensityZone1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-FadeIntensityZone1",
                638,
                0,
                [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                ],
                None,
                16,
                "ALL",
            ),
            "InMix-FadeIntensityZone2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-FadeIntensityZone2",
                638,
                4,
                [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                ],
                None,
                16,
                "ALL",
            ),
            "InMix-FadeIntensityZone3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-FadeIntensityZone3",
                639,
                0,
                [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                ],
                None,
                16,
                "ALL",
            ),
            "InMix-FadeIntensityZone4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMix-FadeIntensityZone4",
                639,
                4,
                [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                ],
                None,
                16,
                "ALL",
            ),
            "InMix-PackType": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/InMix-PackType",
                628,
                None,
                [
                    "Unknown",
                    "inXE",
                    "MasIBC",
                    "MIA",
                    "DJS4",
                    "inClear",
                    "inXM",
                    "K600",
                    "inTerface",
                    "inTouch",
                    "inYT",
                    "Color_Keypad",
                    "inYJ",
                    "MrSteam",
                    "InMix",
                ],
                None,
                None,
                None,
            ),
            "InMix-MemRange": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/InMix-MemRange",
                629,
                0,
                ["8K", "16K", "32K", "64K"],
                None,
                4,
                None,
            ),
            "InMix-FirmwareID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/InMix-FirmwareID", 630, None
            ),
            "InMix-FirmwareRev": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/InMix-FirmwareRev", 632, None
            ),
            "InMix-FirmwareRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/InMix-FirmwareRel", 633, None
            ),
            "InMix-ConfigLib": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/InMix-ConfigLib", 634, None
            ),
            "InMix-StatusLib": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/InMix-StatusLib", 635, None
            ),
            "InMix-ErrorFlag": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Messages/InMix-ErrorFlag", 636, None, None
            ),
            "InMix-ReminderFlag": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Messages/InMix-ReminderFlag", 637, None, None
            ),
        }



"""GeckoLogStruct - A class to manage the LogStruct for 'inClear-32K v3'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoWordStructAccessor,
)


class GeckoLogStruct:
    """Log Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the log struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the log struct class version."""
        return 3

    @property
    def begin(self) -> int:
        """Get the offset start."""
        return 544

    @property
    def end(self) -> int:
        """Get the offset end."""
        return 589

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return []

    @property
    def all_device_keys(self) -> list[str]:
        """Get all device keys."""
        return []

    @property
    def user_demand_keys(self) -> list[str]:
        """Get all user demand keys."""
        return []

    @property
    def error_keys(self) -> list[str]:
        """Get all error keys."""
        return []

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "inClear-32K-Hours": GeckoByteStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-Hours", 544, None
            ),
            "inClear-32K-FlowState": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/All/inClear-32K-FlowState",
                545,
                0,
                ["NO_FLOW", "FLOW"],
                None,
                2,
                None,
            ),
            "inClear-32K-CellPolarity": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/All/inClear-32K-CellPolarity",
                545,
                1,
                ["OUT1", "OUT2"],
                None,
                2,
                None,
            ),
            "inClear-32K-PowerUp": GeckoBoolStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-PowerUp", 545, 2, None
            ),
            "inClear-32K-ResistivityCheck": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/All/inClear-32K-ResistivityCheck",
                545,
                3,
                None,
            ),
            "inClear-32K-MaintenanceQtyByte0": GeckoByteStructAccessor(
                self.struct,
                "LogStructure/All/inClear-32K-MaintenanceQtyByte0",
                546,
                None,
            ),
            "inClear-32K-MaintenanceQtyByte1": GeckoByteStructAccessor(
                self.struct,
                "LogStructure/All/inClear-32K-MaintenanceQtyByte1",
                547,
                None,
            ),
            "inClear-32K-MaintenanceQtyByte2": GeckoByteStructAccessor(
                self.struct,
                "LogStructure/All/inClear-32K-MaintenanceQtyByte2",
                548,
                None,
            ),
            "inClear-32K-MaintenanceQtyByte3": GeckoByteStructAccessor(
                self.struct,
                "LogStructure/All/inClear-32K-MaintenanceQtyByte3",
                549,
                None,
            ),
            "inClear-32K-BoostQtyByte0": GeckoByteStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-BoostQtyByte0", 550, None
            ),
            "inClear-32K-BoostQtyByte1": GeckoByteStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-BoostQtyByte1", 551, None
            ),
            "inClear-32K-BoostQtyByte2": GeckoByteStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-BoostQtyByte2", 552, None
            ),
            "inClear-32K-BoostQtyByte3": GeckoByteStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-BoostQtyByte3", 553, None
            ),
            "inClear-32K-RemainMinutes": GeckoWordStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-RemainMinutes", 554, None
            ),
            "inClear-32K-MaintenanceCurrent": GeckoWordStructAccessor(
                self.struct,
                "LogStructure/All/inClear-32K-MaintenanceCurrent",
                556,
                None,
            ),
            "inClear-32K-BoostCurrent": GeckoWordStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-BoostCurrent", 558, None
            ),
            "inClear-32K-NeededCurrent": GeckoWordStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-NeededCurrent", 560, None
            ),
            "inClear-32K-SupplyVoltage": GeckoWordStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-SupplyVoltage", 562, None
            ),
            "inClear-32K-CellCurrent": GeckoWordStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-CellCurrent", 564, None
            ),
            "inClear-32K-PwmValue": GeckoWordStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-PwmValue", 566, None
            ),
            "inClear-32K-Resistivity": GeckoWordStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-Resistivity", 568, None
            ),
            "inClear-32K-LoSupplyError": GeckoBoolStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-LoSupplyError", 570, 0, None
            ),
            "inClear-32K-LoSaltError": GeckoBoolStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-LoSaltError", 570, 1, None
            ),
            "inClear-32K-LoSaltWarning": GeckoBoolStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-LoSaltWarning", 570, 2, None
            ),
            "inClear-32K-HiSaltWarning": GeckoBoolStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-HiSaltWarning", 570, 3, None
            ),
            "inClear-32K-HiSaltError": GeckoBoolStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-HiSaltError", 570, 5, None
            ),
            "inClear-32K-FloError": GeckoBoolStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-FloError", 570, 6, None
            ),
            "inClear-32K-CompatibilityError": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/All/inClear-32K-CompatibilityError",
                570,
                7,
                None,
            ),
            "inClear-32K-NumberOfResets": GeckoWordStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-NumberOfResets", 571, None
            ),
            "inClear-32K-Action": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/All/inClear-32K-Action",
                573,
                None,
                [
                    "IDLE",
                    "SYSTEM_OFF",
                    "SYSTEM_ON",
                    "BOOST_OFF",
                    "BOOST_ON",
                    "DIAGNOSTIC_OFF",
                    "DIAGNOSTIC_ON",
                ],
                None,
                None,
                "ALL",
            ),
            "inClear-32K-Menu": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/All/inClear-32K-Menu",
                574,
                None,
                [
                    "NORMAL",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "NBR_RESET_MENU",
                    "NBR_RESET_TITLE",
                    "LOWLEVEL_ADJ",
                    "LOWLEVEL_TITLE",
                    "DIAGNOSTIC_MENU",
                ],
                None,
                None,
                None,
            ),
            "inClear-32K-BootID": GeckoWordStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-BootID", 575, None
            ),
            "inClear-32K-BootRev": GeckoByteStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-BootRev", 577, None
            ),
            "inClear-32K-BootRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-BootRel", 578, None
            ),
            "inClear-32K-PackType": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/All/inClear-32K-PackType",
                579,
                None,
                [
                    "Unknown",
                    "inXE",
                    "MasIBC",
                    "MIA",
                    "DJS4",
                    "inClear",
                    "inXM",
                    "K600",
                    "inTerface",
                    "inTouch",
                    "inYT",
                ],
                None,
                None,
                None,
            ),
            "inClear-32K-MemRange": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/All/inClear-32K-MemRange",
                580,
                0,
                ["16K", "32K", "48K", "64K"],
                None,
                4,
                None,
            ),
            "inClear-32K-FirmwareID": GeckoWordStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-FirmwareID", 581, None
            ),
            "inClear-32K-FirmwareRev": GeckoByteStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-FirmwareRev", 583, None
            ),
            "inClear-32K-FirmwareRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-FirmwareRel", 584, None
            ),
            "inClear-32K-ConfigLib": GeckoByteStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-ConfigLib", 585, None
            ),
            "inClear-32K-StatusLib": GeckoByteStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-StatusLib", 586, None
            ),
            "inClear-32K-ResVariation": GeckoByteStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-ResVariation", 587, None
            ),
            "inClear-32K-MinimumFlowTime": GeckoByteStructAccessor(
                self.struct, "LogStructure/All/inClear-32K-MinimumFlowTime", 588, None
            ),
            "inClear-32K-InvalidRemoteFlow": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/All/inClear-32K-InvalidRemoteFlow",
                589,
                0,
                ["USE_PACK_FLOW", "NOT_USE_PACK_FLOW"],
                None,
                2,
                "ALL",
            ),
            "inClear-32K-DoNotForceFilter": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/All/inClear-32K-DoNotForceFilter",
                589,
                1,
                ["FORCE_FILTER", "NOT_FORCE_FILTER"],
                None,
                2,
                "ALL",
            ),
            "inClear-32K-PSNotInstalled": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/All/inClear-32K-PSNotInstalled",
                589,
                2,
                ["WITH_PS", "WITHOUT_PS"],
                None,
                2,
                None,
            ),
        }



"""Shared command functionality."""

# ruff: noqa: T201

import fnmatch
import glob
import logging
import os
from typing import TYPE_CHECKING

from geckolib.async_taskman import GeckoAsyncTaskMan
from geckolib.const import GeckoConstants

from .async_command import AsyncCmd

if TYPE_CHECKING:
    from geckolib.driver import GeckoStructAccessor
    from geckolib.driver.async_spastruct import GeckoAsyncStructure

_LOGGER = logging.getLogger(__name__)

LICENSE = """
#
#   Copyright (C) 2020, Gazoodle (https://github.com/gazoodle)
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""


class GeckoCmd(AsyncCmd):
    """
    Shared command processor.

    Implements shared functionality between the Shell and the Simulator

    This version is now async.
    """

    BANNER = None

    @classmethod
    def run(cls, first_commands: list[str] | None = None) -> None:
        """Run command loop."""
        with cls() as cmd:  # type: ignore  # noqa: PGH003
            if first_commands is not None:
                for command in first_commands:
                    cmd.push_command(command)
            cmd.cmdloop()

    def __init__(self, taskman: GeckoAsyncTaskMan) -> None:
        """Initialize the command class."""
        super().__init__()
        self._taskman = taskman

        if self.BANNER is not None:
            print(self.BANNER)

        self.stream_logger = None
        self.file_logger = None
        self._init_logging()

        self.structure: GeckoAsyncStructure | None = None
        self.running_set_command: bool = False

    def push_command(self, cmd: str) -> None:
        """Add a command to the queue."""
        self.cmdqueue.append(cmd)

    def do_loglevel(self, arg: str) -> None:
        """Set the logging level : loglevel <ERROR|WARNING|INFO|DEBUG>."""
        for handler in logging.getLogger().handlers:
            handler.setLevel(arg)
        self._set_root_log_level()

    def do_logfile(self, arg: str) -> None:
        """Add a file logger to the system : logfile <filename>."""
        if self.file_logger is not None:
            print("There is already a file logger installed")
            return
        self.file_logger = logging.FileHandler(arg)
        self.file_logger.setLevel(logging.DEBUG)
        self.file_logger.setFormatter(
            logging.Formatter("%(asctime)s %(name)s %(levelname)s %(message)s")
        )
        logging.getLogger().addHandler(self.file_logger)
        self._set_root_log_level()

    def _init_logging(self) -> None:
        self.stream_logger = logging.StreamHandler()
        self.stream_logger.setLevel(logging.WARNING)
        self.stream_logger.setFormatter(
            logging.Formatter("LOG> %(levelname)s %(message)s")
        )
        logging.getLogger().addHandler(self.stream_logger)
        self._set_root_log_level()

    def _set_root_log_level(self) -> None:
        # Set root log level
        logging.getLogger().setLevel(
            min(handler.level for handler in logging.getLogger().handlers)
        )

    def do_license(self, _arg: str) -> None:
        """Display the license details : license."""
        print(LICENSE)

    def emptyline(self) -> None:
        """Call when an empty line is entered in response to the prompt."""

    def do_get(self, arg: str) -> None:
        """
        Get the value of the specified spa pack structure element.

        Usage: get <Element>
        """
        if self.structure is None:
            print("No access to the spa structure")
            return
        try:
            print(self.structure.accessors[arg].diag_info())
        except Exception:  # pylint: disable=broad-except
            _LOGGER.exception("Exception getting '%s'", arg)

    def _accessor_list(self, pattern: str) -> list[str]:
        if self.structure is None:
            return []
        keys = self.structure.accessors.keys()
        if not pattern:
            pattern = ""
        if not pattern.endswith("?"):
            pattern = pattern + "*"
        return fnmatch.filter(keys, pattern)

    def complete_get(
        self, text: str, _line: int, _start_idx: int, _end_idx: int
    ) -> list[str]:
        """Complete the get command."""
        return self._accessor_list(text)

    def do_accessors(self, pattern: str) -> None:
        """
        Display the values from the accessors.

        Usage: accessors [pattern] where pattern is case sensitive
        text with * and ? wildcards
        """
        if self.structure is None:
            print("No access to the spa structure")
            return
        print("Accessors")
        print("=========")
        print()
        for key in self._accessor_list(pattern):
            print(f"   {self.structure.accessors[key]!r}")
        print()

    def complete_accessors(
        self, text: str, _line: int, _start_idx: int, _end_idx: int
    ) -> list[str]:
        """Complete the accessors command."""
        return self._accessor_list(text)

    def do_peek(self, arg: str) -> None:
        """
        Get the byte value from the structure at the specified position.

        Usage: peek <pos>
        """
        if self.structure is None:
            print("No access to the spa structure")
            return
        try:
            pos = int(arg)
            print(
                f"Byte at {pos} = {self.structure.status_block[pos]} (\\x{self.structure.status_block[pos]:x})"  # noqa: E501
            )
        except Exception:  # pylint: disable=broad-except
            _LOGGER.exception("Exception peeking at '%s'", arg)

    async def do_set(self, arg: str) -> None:
        """
        Set the value of the specified spa pack structure element.

        Usage: set <Element>=<value>
        """
        if self.structure is None:
            print("No access to the spa structure")
            return
        self.running_set_command = True
        try:
            key, val = arg.split("=")
            await self.structure.accessors[key].async_set_value(val)
        except Exception:  # pylint: disable=broad-except
            _LOGGER.exception("Exception handling 'set %s'", arg)
        finally:
            self.running_set_command = False

    def complete_set(
        self, text: str, _line: int, _start_idx: int, _end_idx: int
    ) -> list[str]:
        """Complete the set command."""
        if "=" not in text:
            return self._accessor_list(text)
        key, val = text.split("=")
        if self.structure is None:
            return []
        if key not in self.structure.accessors:
            return []
        accessor: GeckoStructAccessor = self.structure.accessors[key]
        if accessor.accessor_type != GeckoConstants.SPA_PACK_STRUCT_ENUM_TYPE:
            return []
        return [f"{key}={item}" for item in accessor.items if item.startswith(val)]

    def do_hexdump(self, _arg: str) -> None:
        """Hex dump of the raw structure status block."""
        if self.structure is None:
            print("No access to the spa structure")
            return

        width = 16
        data = self.structure.status_block

        lines = []
        for i in range(0, len(data), width):
            chunk = data[i : i + width]
            hex_values = " ".join(f"{b:02x}" for b in chunk)
            # Split hex values into two parts with an extra space in the middle
            left_hex = hex_values[:23]  # 8 hex octets + 7 spaces = 23 characters
            right_hex = hex_values[23:]  # The rest
            formatted_hex = f"{left_hex} {right_hex}"
            printable = "".join((chr(b) if 32 <= b < 127 else ".") for b in chunk)  # noqa: PLR2004
            lines.append(f"{i:08x}  {formatted_hex:<{width * 3 + 1}}  {printable}")

        print("Status block")
        print("\n".join(lines))

    def do_snapshot(self, arg: str) -> None:
        """
        Take a snapshot of the spa data structure.

        Supply an option description to help identify
        the state of the spa at this time.

        usage: snapshot <desc>
        """
        data = self.get_snapshot_data()
        if arg:
            data["Description"] = arg
        _LOGGER.info(f"SNAPSHOT ========{data}========")  # noqa: G004
        print(data)

    def _complete_path(self, path: str) -> list[str]:
        if os.path.isdir(path):  # noqa: PTH112
            return glob.glob(os.path.join(path, "*"))  # noqa: PTH118, PTH207
        return glob.glob(path + "*")  # noqa: PTH207

    def do_pack(self, _args: str) -> None:
        """List pack info."""
        print(f"Outputs {self.structure.all_outputs}")
        print(f"Connections: {self.structure.connections}")
        print(f"All Devices: {self.structure.all_devices}")
        print(f"Error Keys: {self.structure.error_keys}")
        print(f"User Demands: {self.structure.user_demands}")



"""Unit tests for the UDP protocol handlers."""  # noqa: INP001

# ruff: noqa: E501

import unittest
import unittest.mock

import pytest
from context import (
    GeckoAsyncPartialStatusBlockProtocolHandler,
    GeckoStatusBlockProtocolHandler,
    GeckoUdpProtocolHandler,
)

from geckolib.driver.async_udp_protocol import GeckoAsyncUdpProtocol

PARMS = (1, 2, b"SRCID", b"DESTID")


class MockProtocol(GeckoAsyncUdpProtocol):
    """Mock protocol class."""

    def __init__(self) -> None:
        """Initialiale the mock protocol class."""
        self.sendbytes = None

    def queue_send(
        self,
        protocol_handler: GeckoUdpProtocolHandler,
        _destination: tuple | None = None,
    ) -> None:
        self.sendbytes = protocol_handler.send_bytes

    def get_and_increment_sequence_counter(self, *, command: bool = False) -> int:
        if command:
            return 192
        return 1


class TestGeckoStatusBlockHandler(unittest.TestCase):
    """Status block tests."""

    def test_send_construct_request(self) -> None:
        handler = GeckoStatusBlockProtocolHandler.request(1, 0, 637, parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>STATU\x01\x00\x00\x02\x7d</DATAS></PACKT>",
        )

    def test_send_construct_response(self) -> None:
        handler = GeckoStatusBlockProtocolHandler.response(
            3, 4, b"\x01\x02\x03\x04", parms=PARMS
        )
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>STATV\x03\x04\x04\x01\x02\x03\x04</DATAS></PACKT>",
        )

    def test_recv_can_handle(self) -> None:
        handler = GeckoStatusBlockProtocolHandler()
        self.assertTrue(handler.can_handle(b"STATU", PARMS))
        self.assertTrue(handler.can_handle(b"STATV", PARMS))
        self.assertFalse(handler.can_handle(b"OTHER", PARMS))

    def test_recv_partial_can_handle(self) -> None:
        protocol = MockProtocol()
        handler = GeckoAsyncPartialStatusBlockProtocolHandler(protocol)
        self.assertTrue(handler.can_handle(b"STATP", PARMS))
        self.assertTrue(handler.can_handle(b"STATQ", PARMS))
        self.assertFalse(handler.can_handle(b"OTHER", PARMS))
        self.assertIsNone(protocol.sendbytes)

    def test_recv_handle_request(self) -> None:
        handler = GeckoStatusBlockProtocolHandler()
        handler.handle(b"STATU\x01\x00\x00\x02\x7d", PARMS)
        self.assertFalse(handler.should_remove_handler)
        self.assertEqual(handler.sequence, 1)
        self.assertEqual(handler.start, 0)
        self.assertEqual(handler.length, 637)

    def test_recv_handle_response(self) -> None:
        handler = GeckoStatusBlockProtocolHandler()
        handler.handle(b"STATV\x03\x04\x04\x01\x02\x03\x04", PARMS)
        self.assertFalse(handler.should_remove_handler)
        self.assertEqual(handler.sequence, 3)
        self.assertEqual(handler.next, 4)
        self.assertEqual(handler.length, 4)
        self.assertEqual(handler.data, b"\x01\x02\x03\x04")

    def test_recv_handle_response_final(self) -> None:
        handler = GeckoStatusBlockProtocolHandler()
        handler.handle(b"STATV\x03\x00\x04\x01\x02\x03\x04", PARMS)
        self.assertFalse(handler.should_remove_handler)
        self.assertEqual(handler.sequence, 3)
        self.assertEqual(handler.next, 0)
        self.assertEqual(handler.length, 4)
        self.assertEqual(handler.data, b"\x01\x02\x03\x04")

    def test_send_partial(self) -> None:
        protocol = MockProtocol()
        handler = GeckoAsyncPartialStatusBlockProtocolHandler.report_changes(
            protocol, [(365, b"\x03\x84"), (366, b"\x84\x0c")], parms=PARMS
        )
        self.assertFalse(handler.should_remove_handler)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN><DATAS>STATP\x02\x01\x6d\x03\x84\x01\x6e\x84\x0c</DATAS></PACKT>",
        )

    def test_send_partial_three(self) -> None:
        protocol = MockProtocol()
        handler = GeckoAsyncPartialStatusBlockProtocolHandler.report_changes(
            protocol,
            [(365, b"\x03\x84"), (366, b"\x84\x0c"), (367, b"\x01\x02")],
            parms=PARMS,
        )
        self.assertFalse(handler.should_remove_handler)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN><DATAS>STATP\x03\x01\x6d\x03\x84\x01\x6e\x84\x0c\x01\x6f\x01\x02</DATAS></PACKT>",
        )

    @unittest.expectedFailure
    def test_send_partial_not_two_bytes(self) -> None:
        protocol = MockProtocol()
        handler = GeckoAsyncPartialStatusBlockProtocolHandler.report_changes(
            protocol,
            [(367, b"\x01")],
            parms=PARMS,
        )
        self.assertFalse(handler.should_remove_handler)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN><DATAS>STATP\x03\x01\x6d\x03\x84\x01\x6e\x84\x0c\x01\x6f\x01\x02</DATAS></PACKT>",
        )


class TestGeckoAsyncStatusBlockHandler:
    """Status block tests."""

    @pytest.mark.asyncio
    async def test_recv_handle_partial(self) -> None:
        protocol = MockProtocol()
        handler = GeckoAsyncPartialStatusBlockProtocolHandler(protocol)
        await handler.async_handle(b"STATV\x02\x01\x6d\x03\x84\x01\x6e\x84\x0c", PARMS)
        assert not handler.should_remove_handler
        assert handler.changes == [(365, b"\x03\x84"), (366, b"\x84\x0c")]
        assert (
            protocol.sendbytes
            == b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN><DATAS>STATQ\xc0</DATAS></PACKT>"
        )

    @pytest.mark.asyncio
    async def test_change_unpack_more(self) -> None:
        protocol = MockProtocol()
        handler = GeckoAsyncPartialStatusBlockProtocolHandler(protocol)
        await handler.async_handle(
            b"STATV\x05\x00\x01\x02\x02\x00\x02\x04\x04\x00\x03\x06\x06\x00\x04\x08\x08\x00\x05\x0a\x0a",
            PARMS,
        )
        assert handler.changes == [
            (1, b"\x02\x02"),
            (2, b"\x04\x04"),
            (3, b"\x06\x06"),
            (4, b"\x08\x08"),
            (5, b"\x0a\x0a"),
        ]
        assert (
            protocol.sendbytes
            == b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN><DATAS>STATQ\xc0</DATAS></PACKT>"
        )


if __name__ == "__main__":
    unittest.main()



"""GeckoLogStruct - A class to manage the LogStruct for 'InMixExtended v2'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoWordStructAccessor,
)


class GeckoLogStruct:
    """Log Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the log struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the log struct class version."""
        return 2

    @property
    def begin(self) -> int:
        """Get the offset start."""
        return 704

    @property
    def end(self) -> int:
        """Get the offset end."""
        return 741

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return []

    @property
    def all_device_keys(self) -> list[str]:
        """Get all device keys."""
        return []

    @property
    def user_demand_keys(self) -> list[str]:
        """Get all user demand keys."""
        return []

    @property
    def error_keys(self) -> list[str]:
        """Get all error keys."""
        return []

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "InMixExtended-Synchro1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-Synchro1",
                704,
                None,
                ["ZONE1", "ZONE2", "ZONE3", "ZONE4"],
                None,
                None,
                "ALL",
            ),
            "InMixExtended-Mode1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-Mode1",
                705,
                None,
                [
                    "STATIC",
                    "VERY_SLOW_FADE",
                    "SLOW_FADE",
                    "BENOIT1",
                    "BENOIT2",
                    "RGB",
                    "DELAY_ADJUSTABLE_FADE_MODE",
                    "LIMITED_COLOR_FADE_MODE",
                ],
                None,
                None,
                "ALL",
            ),
            "InMixExtended-Color1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-Color1",
                706,
                None,
                [
                    "NO_COLOR",
                    "WHITE",
                    "RED",
                    "GREEN",
                    "BLUE",
                    "PURPLE",
                    "YELLOW",
                    "CYAN",
                    "ORANGE",
                    "MAGENTA",
                    "ROYAL_BLUE",
                ],
                None,
                None,
                "ALL",
            ),
            "InMixExtended-Speed1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-Speed1",
                707,
                None,
                [
                    "PRESET",
                    "LOW",
                    "MEDIUM",
                    "HIGH",
                    "PAUSE",
                    "FADE_ADJUST_DELAY0",
                    "FADE_ADJUST_DELAY1",
                    "FADE_ADJUST_DELAY2",
                    "FADE_ADJUST_DELAY3",
                    "FADE_ADJUST_DELAY4",
                    "FADE_ADJUST_DELAY5",
                ],
                None,
                None,
                "ALL",
            ),
            "InMixExtended-RedLevel1": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMixExtended-RedLevel1", 708, "ALL"
            ),
            "InMixExtended-GreenLevel1": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMixExtended-GreenLevel1", 709, "ALL"
            ),
            "InMixExtended-BlueLevel1": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMixExtended-BlueLevel1", 710, "ALL"
            ),
            "InMixExtended-Synchro2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-Synchro2",
                711,
                None,
                ["ZONE1", "ZONE2", "ZONE3", "ZONE4"],
                None,
                None,
                "ALL",
            ),
            "InMixExtended-Mode2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-Mode2",
                712,
                None,
                [
                    "STATIC",
                    "VERY_SLOW_FADE",
                    "SLOW_FADE",
                    "BENOIT1",
                    "BENOIT2",
                    "RGB",
                    "DELAY_ADJUSTABLE_FADE_MODE",
                    "LIMITED_COLOR_FADE_MODE",
                ],
                None,
                None,
                "ALL",
            ),
            "InMixExtended-Color2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-Color2",
                713,
                None,
                [
                    "NO_COLOR",
                    "WHITE",
                    "RED",
                    "GREEN",
                    "BLUE",
                    "PURPLE",
                    "YELLOW",
                    "CYAN",
                    "ORANGE",
                    "MAGENTA",
                    "ROYAL_BLUE",
                ],
                None,
                None,
                "ALL",
            ),
            "InMixExtended-Speed2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-Speed2",
                714,
                None,
                [
                    "PRESET",
                    "LOW",
                    "MEDIUM",
                    "HIGH",
                    "PAUSE",
                    "FADE_ADJUST_DELAY0",
                    "FADE_ADJUST_DELAY1",
                    "FADE_ADJUST_DELAY2",
                    "FADE_ADJUST_DELAY3",
                    "FADE_ADJUST_DELAY4",
                    "FADE_ADJUST_DELAY5",
                ],
                None,
                None,
                "ALL",
            ),
            "InMixExtended-RedLevel2": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMixExtended-RedLevel2", 715, "ALL"
            ),
            "InMixExtended-GreenLevel2": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMixExtended-GreenLevel2", 716, "ALL"
            ),
            "InMixExtended-BlueLevel2": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMixExtended-BlueLevel2", 717, "ALL"
            ),
            "InMixExtended-Synchro3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-Synchro3",
                718,
                None,
                ["ZONE1", "ZONE2", "ZONE3", "ZONE4"],
                None,
                None,
                "ALL",
            ),
            "InMixExtended-Mode3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-Mode3",
                719,
                None,
                [
                    "STATIC",
                    "VERY_SLOW_FADE",
                    "SLOW_FADE",
                    "BENOIT1",
                    "BENOIT2",
                    "RGB",
                    "DELAY_ADJUSTABLE_FADE_MODE",
                    "LIMITED_COLOR_FADE_MODE",
                ],
                None,
                None,
                "ALL",
            ),
            "InMixExtended-Color3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-Color3",
                720,
                None,
                [
                    "NO_COLOR",
                    "WHITE",
                    "RED",
                    "GREEN",
                    "BLUE",
                    "PURPLE",
                    "YELLOW",
                    "CYAN",
                    "ORANGE",
                    "MAGENTA",
                    "ROYAL_BLUE",
                ],
                None,
                None,
                "ALL",
            ),
            "InMixExtended-Speed3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-Speed3",
                721,
                None,
                [
                    "PRESET",
                    "LOW",
                    "MEDIUM",
                    "HIGH",
                    "PAUSE",
                    "FADE_ADJUST_DELAY0",
                    "FADE_ADJUST_DELAY1",
                    "FADE_ADJUST_DELAY2",
                    "FADE_ADJUST_DELAY3",
                    "FADE_ADJUST_DELAY4",
                    "FADE_ADJUST_DELAY5",
                ],
                None,
                None,
                "ALL",
            ),
            "InMixExtended-RedLevel3": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMixExtended-RedLevel3", 722, "ALL"
            ),
            "InMixExtended-GreenLevel3": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMixExtended-GreenLevel3", 723, "ALL"
            ),
            "InMixExtended-BlueLevel3": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMixExtended-BlueLevel3", 724, "ALL"
            ),
            "InMixExtended-Synchro4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-Synchro4",
                725,
                None,
                ["ZONE1", "ZONE2", "ZONE3", "ZONE4"],
                None,
                None,
                "ALL",
            ),
            "InMixExtended-Mode4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-Mode4",
                726,
                None,
                [
                    "STATIC",
                    "VERY_SLOW_FADE",
                    "SLOW_FADE",
                    "BENOIT1",
                    "BENOIT2",
                    "RGB",
                    "DELAY_ADJUSTABLE_FADE_MODE",
                    "LIMITED_COLOR_FADE_MODE",
                ],
                None,
                None,
                "ALL",
            ),
            "InMixExtended-Color4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-Color4",
                727,
                None,
                [
                    "NO_COLOR",
                    "WHITE",
                    "RED",
                    "GREEN",
                    "BLUE",
                    "PURPLE",
                    "YELLOW",
                    "CYAN",
                    "ORANGE",
                    "MAGENTA",
                    "ROYAL_BLUE",
                ],
                None,
                None,
                "ALL",
            ),
            "InMixExtended-Speed4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-Speed4",
                728,
                None,
                [
                    "PRESET",
                    "LOW",
                    "MEDIUM",
                    "HIGH",
                    "PAUSE",
                    "FADE_ADJUST_DELAY0",
                    "FADE_ADJUST_DELAY1",
                    "FADE_ADJUST_DELAY2",
                    "FADE_ADJUST_DELAY3",
                    "FADE_ADJUST_DELAY4",
                    "FADE_ADJUST_DELAY5",
                ],
                None,
                None,
                "ALL",
            ),
            "InMixExtended-RedLevel4": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMixExtended-RedLevel4", 729, "ALL"
            ),
            "InMixExtended-GreenLevel4": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMixExtended-GreenLevel4", 730, "ALL"
            ),
            "InMixExtended-BlueLevel4": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/InMixExtended-BlueLevel4", 731, "ALL"
            ),
            "InMixExtended-FadeIntensityZone1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-FadeIntensityZone1",
                732,
                0,
                [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                ],
                None,
                16,
                "ALL",
            ),
            "InMixExtended-FadeIntensityZone2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-FadeIntensityZone2",
                732,
                4,
                [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                ],
                None,
                16,
                "ALL",
            ),
            "InMixExtended-FadeIntensityZone3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-FadeIntensityZone3",
                733,
                0,
                [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                ],
                None,
                16,
                "ALL",
            ),
            "InMixExtended-FadeIntensityZone4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/InMixExtended-FadeIntensityZone4",
                733,
                4,
                [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                ],
                None,
                16,
                "ALL",
            ),
            "InMixExtended-PackType": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/InMixExtended-PackType",
                732,
                None,
                [
                    "Unknown",
                    "inXE",
                    "MasIBC",
                    "MIA",
                    "DJS4",
                    "inClear",
                    "inXM",
                    "K600",
                    "inTerface",
                    "inTouch",
                    "inYT",
                    "Color_Keypad",
                    "inYJ",
                    "MrSteam",
                    "InMix",
                ],
                None,
                None,
                None,
            ),
            "InMixExtended-MemRange": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/InMixExtended-MemRange",
                733,
                0,
                ["8K", "16K", "32K", "64K"],
                None,
                4,
                None,
            ),
            "InMixExtended-FirmwareID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/InMixExtended-FirmwareID", 734, None
            ),
            "InMixExtended-FirmwareRev": GeckoByteStructAccessor(
                self.struct,
                "LogStructure/PackInfo/InMixExtended-FirmwareRev",
                736,
                None,
            ),
            "InMixExtended-FirmwareRel": GeckoByteStructAccessor(
                self.struct,
                "LogStructure/PackInfo/InMixExtended-FirmwareRel",
                737,
                None,
            ),
            "InMixExtended-ConfigLib": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/InMixExtended-ConfigLib", 738, None
            ),
            "InMixExtended-StatusLib": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/InMixExtended-StatusLib", 739, None
            ),
            "InMixExtended-ErrorFlag": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/Messages/InMixExtended-ErrorFlag",
                740,
                None,
                None,
            ),
            "InMixExtended-ReminderFlag": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/Messages/InMixExtended-ReminderFlag",
                741,
                None,
                None,
            ),
        }



"""Gecko inMix accessory class."""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any

from geckolib.automation.light import GeckoLight
from geckolib.automation.power import GeckoPower
from geckolib.automation.select import GeckoSelect

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade
    from geckolib.driver.accessor import (
        GeckoByteStructAccessor,
        GeckoEnumStructAccessor,
    )

_LOGGER = logging.getLogger(__name__)

ATTR_RGBCOLOR = "rgb_color"
ATTR_BRIGHTNESS = "brightness"
ZONE_KEY_NONE = "None"

"""
    Notes:

    InMix-Color[n]: If not NO_COLOR, then InMix-Mode[n] must be STATIC


"""


class GeckoInMixZone(GeckoLight):
    """Class for an inMix light."""

    def __init__(self, facade: GeckoAsyncFacade, inmix: GeckoInMix, zone: int) -> None:
        """Initialize the Li light."""
        super().__init__(facade, f"Zone {zone}", f"InMixZone{zone}")
        if zone <= inmix.number_of_zones:
            self.set_availability(is_available=True)
        else:
            return

        self._synchro_accessor: GeckoEnumStructAccessor = self.facade.spa.accessors[
            f"InMix-Synchro{zone}"
        ]
        self._mode_accessor: GeckoEnumStructAccessor = self.facade.spa.accessors[
            f"InMix-Mode{zone}"
        ]
        self._color_accessor: GeckoEnumStructAccessor = self.facade.spa.accessors[
            f"InMix-Color{zone}"
        ]
        self._speed_accessor: GeckoEnumStructAccessor = self.facade.spa.accessors[
            f"InMix-Speed{zone}"
        ]
        self._red_accessor: GeckoByteStructAccessor = self.facade.spa.accessors[
            f"InMix-RedLevel{zone}"
        ]
        self._green_accessor: GeckoByteStructAccessor = self.facade.spa.accessors[
            f"InMix-GreenLevel{zone}"
        ]
        self._blue_accessor: GeckoByteStructAccessor = self.facade.spa.accessors[
            f"InMix-BlueLevel{zone}"
        ]

        # Set accessors for direct struct update, and watch them
        for accessor in [
            self._synchro_accessor,
            self._mode_accessor,
            self._color_accessor,
            self._speed_accessor,
            self._red_accessor,
            self._green_accessor,
            self._blue_accessor,
        ]:
            accessor.direct_update = True
            accessor.watch(self._on_change)

        self._brightness = 255
        self._zone = zone
        self._get_inmix()
        self._ignore_changes = False

    @property
    def is_on(self) -> bool:
        """Determine if the light is on or not."""
        return self._mode_accessor.value != "STATIC"

    @property
    def rgb_color(self) -> tuple[int, int, int] | None:
        """Get the RGB colour."""
        if self.is_on:
            return self._rgb
        return None

    @property
    def brightness(self) -> int | None:
        """Get the zone brightness."""
        if self.is_on:
            return self._brightness
        return None

    @property
    def zone_key(self) -> str:
        """Get the zone key."""
        return f"ZONE{self._zone}"

    @property
    def synchro(self) -> str:
        """Get this zone sync value."""
        return self._synchro_accessor.value

    async def set_synchro(self, zone: GeckoInMixZone | None) -> None:
        """Set the zone sync value."""
        if zone is None:
            await self._synchro_accessor.async_set_value(self.zone_key)
            return
        if self.zone_key == zone.zone_key:
            return
        if zone.is_on:
            await self.async_turn_on(
                **{ATTR_BRIGHTNESS: zone.brightness, ATTR_RGBCOLOR: zone.rgb_color}
            )
            return
        await self.async_turn_off()

    @property
    def is_synchro(self) -> bool:
        """Is this zone synchro with another."""
        return self.synchro != self.zone_key

    @property
    def state(self) -> Any:
        """Get the state of the light."""
        if self.is_on:
            return f"RGB({self._rgb}) {int(self._brightness * 100 / 255)}%"
        return "OFF"

    async def _set_rgb(self, rgb: tuple[int, int, int], brightness: int) -> None:
        _LOGGER.debug("Set RGB %s %s", rgb, brightness)
        self._rgb = rgb
        self._brightness = brightness
        await self._set_inmix()

    async def _set_brightness(self, brightness: int) -> None:
        _LOGGER.debug("Set brightness %d", brightness)
        self._brightness = brightness
        await self._set_inmix()

    async def _set_inmix(self) -> None:
        self._ignore_changes = True
        await self._mode_accessor.async_set_value("RGB")
        rgb = self._rgb
        br = self._brightness
        await self._red_accessor.async_set_value(int(rgb[0] * br / 255))
        await self._green_accessor.async_set_value(int(rgb[1] * br / 255))
        await self._blue_accessor.async_set_value(int(rgb[2] * br / 255))
        self._ignore_changes = False
        self._on_change()

    async def async_turn_on(self, **kwargs: Any) -> None:
        """Turn the light ON, but does nothing if it is already ON."""
        _LOGGER.info("On %s", kwargs)
        brightness = kwargs.get(ATTR_BRIGHTNESS, self._brightness)
        if ATTR_RGBCOLOR in kwargs:
            rgb = kwargs.get(ATTR_RGBCOLOR)
            await self._set_rgb(rgb, brightness)
            return
        if ATTR_BRIGHTNESS in kwargs:
            await self._set_brightness(brightness)
            return
        if self.is_on:
            return
        await self._mode_accessor.async_set_value("RGB")

    async def async_turn_off(self, **_kwargs: Any) -> None:
        """Turn the light OFF, but does nothing if it is already OFF."""
        if not self.is_on:
            return
        await self._mode_accessor.async_set_value("STATIC")

    def _get_inmix(self) -> None:
        rgb = (
            self._red_accessor.value,
            self._green_accessor.value,
            self._blue_accessor.value,
        )
        _LOGGER.debug("get_inmix: Got %s", rgb)
        self._brightness: int = max(1, *rgb)
        # Scale the colours from the apparent brightness
        self._rgb = (
            int(255 * rgb[0] / self._brightness),
            int(255 * rgb[1] / self._brightness),
            int(255 * rgb[2] / self._brightness),
        )
        _LOGGER.debug("scaled %s", self._rgb)

    def _on_change(
        self, sender: Any = None, old_value: Any = None, new_value: Any = None
    ) -> None:
        if self._ignore_changes:
            return
        self._get_inmix()
        super()._on_change(sender, old_value, new_value)


class GeckoInMixSynchro(GeckoSelect):
    """Class for the synchro setting."""

    def __init__(self, facade: GeckoAsyncFacade, inmix: GeckoInMix) -> None:
        """Initialize the sync class."""
        super().__init__(facade, "Zone Syncronization", "INMIXSYNC")
        if inmix.number_of_zones == 1:
            return
        self.set_availability(is_available=True)

        # Set of mappings of constants to UI options. There must be at
        # least 2 zones
        mappings = {
            ZONE_KEY_NONE: "None",
            inmix.zone_1.zone_key: "With Zone 1",
            inmix.zone_2.zone_key: "With Zone 2",
        }
        if inmix.number_of_zones == 3:  # noqa: PLR2004
            mappings[inmix.zone_3.zone_key] = "With Zone 3"
        self.set_mapping(mappings)

        self._inmix = inmix
        self._lookup: dict = {ZONE_KEY_NONE: None}
        for zone in inmix.zones:
            self._lookup[zone.zone_key] = zone
            zone.watch(self._on_zone_change)

        self._state = ZONE_KEY_NONE
        self._ignore_changes = False

    def _on_zone_change(
        self, _sender: Any = None, _old_value: Any = None, _new_value: Any = None
    ) -> None:
        if self._ignore_changes:
            return
        try:
            for zone in self._inmix.zones:
                if zone.is_synchro:
                    self._state = zone.synchro
                    return
            self._state = ZONE_KEY_NONE
        finally:
            self._on_change(None, None, None)

    @property
    def state(self) -> str:
        """Get the current state via the mapping."""
        return self.mapping[self._state]

    async def async_set_state(self, new_state: str) -> None:
        """Set the state of the select entity."""
        if new_state in self.reverse:
            new_state = self.reverse[new_state]
        self._state = new_state
        try:
            self._ignore_changes = True
            for zone in self._inmix.zones:
                await zone.set_synchro(self._lookup[self._state])
        finally:
            self._ignore_changes = False

    @property
    def states(self) -> list[str]:
        """Get the possible states."""
        return list(self.mapping.values())


class GeckoInMix(GeckoPower):
    """Gecko inMix support class."""

    def __init__(self, facade: GeckoAsyncFacade) -> None:
        """Initialize the inMix class."""
        super().__init__(facade, "inMix", "INMIX")
        if "InMix-PackType" not in facade.spa.accessors:
            return

        self.struct_begin = facade.spa.struct.inmix_log_class.begin

        self.number_of_zones: int = int(
            facade.spa.accessors["InMix-NumberOfZones"].value
        )
        _LOGGER.debug("Spa has an inMix accessory with %d zones", self.number_of_zones)

        self.set_availability(is_available=True)

        self.zone_1 = GeckoInMixZone(facade, self, 1)
        self.zone_2 = GeckoInMixZone(facade, self, 2)
        self.zone_3 = GeckoInMixZone(facade, self, 3)
        self.syncro = GeckoInMixSynchro(facade, self)

        #
        #   If one zone, then no synchro available
        #   Each zone either preset or user selected
        #   RGB 0-255 per normal
        #

        self.zone_1.watch(self._on_change)
        self.zone_2.watch(self._on_change)
        self.zone_3.watch(self._on_change)
        self.syncro.watch(self._on_change)

        # Start task for
        # facade.taskmanager.add_task(self._in
        facade.taskmanager.add_task(self._inmix_update(), "inix Update", "FACADE")
        self._update_event = asyncio.Event()

    async def _inmix_update(self) -> None:
        _LOGGER.debug("inMix update task started")
        try:
            while True:
                await self._update_event.wait()

                _LOGGER.debug("Perform struct update")
                block_size = len(self.zones) * 7
                data = self.facade.spa.struct.status_block[
                    self.struct_begin : self.struct_begin + block_size
                ]
                await self.facade.spa.async_on_set_value(
                    self.struct_begin, block_size, data
                )
                self._update_event.clear()

        except asyncio.CancelledError:
            _LOGGER.debug("Facade update loop cancelled")
            raise
        except Exception:
            _LOGGER.exception("Facade update loop caught execption")
            raise

    @property
    def zones(self) -> list[GeckoInMixZone]:
        """Get the available zones."""
        return [
            zone
            for zone in [self.zone_1, self.zone_2, self.zone_3]
            if zone.is_available
        ]

    def _on_change(
        self, sender: Any = None, old_value: Any = None, new_value: Any = None
    ) -> None:
        self._update_event.set()
        return super()._on_change(sender, old_value, new_value)



"""GeckoConfigStruct - A class to manage the ConfigStruct for 'MrSteam v1'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoWordStructAccessor,
)


class GeckoConfigStruct:
    """Config Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the config struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the config struct class version."""
        return 1

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return []

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "Prog1Setpoint": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/All/Prog1Setpoint", 0, "ALL"
            ),
            "Prog1Runtime": GeckoWordStructAccessor(
                self.struct, "ConfigStructure/All/Prog1Runtime", 1, "ALL"
            ),
            "Prog1Aroma": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/All/Prog1Aroma",
                3,
                None,
                ["OFF", "ON"],
                None,
                None,
                "ALL",
            ),
            "Prog2Setpoint": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/All/Prog2Setpoint", 4, "ALL"
            ),
            "Prog2Runtime": GeckoWordStructAccessor(
                self.struct, "ConfigStructure/All/Prog2Runtime", 5, "ALL"
            ),
            "Prog2Aroma": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/All/Prog2Aroma",
                7,
                None,
                ["OFF", "ON"],
                None,
                None,
                "ALL",
            ),
            "TempUnits": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/All/TempUnits",
                8,
                None,
                ["F", "C"],
                None,
                None,
                "ALL",
            ),
        }



"""GeckoAsyncTaskMan is a wrapper on AsyncTasks with some shared properties."""

from .async_tasks import AsyncTasks


class GeckoAsyncTaskMan(AsyncTasks):
    """Async tasks with spa name and id."""

    unique_id = ""
    spa_name = ""

    def __init__(self) -> None:
        """Initialize the async task manager class."""
        super().__init__()



"""Unit tests for the UDP protocol handlers."""  # noqa: INP001

# ruff: noqa: E501

import unittest
import unittest.mock

from context import (
    GeckoGetWatercareModeProtocolHandler,
    GeckoGetWatercareScheduleListProtocolHandler,
    GeckoSetWatercareModeProtocolHandler,
    GeckoWatercareScheduleManager,
)

PARMS = (1, 2, b"SRCID", b"DESTID")


class TestGeckoGetWatercareModeHandler(unittest.TestCase):
    """Get Watercare handler tests."""

    def test_send_construct_get(self) -> None:
        handler = GeckoGetWatercareModeProtocolHandler.get(1, parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>GETWC\x01</DATAS></PACKT>",
        )

    def test_send_construct_response(self) -> None:
        handler = GeckoGetWatercareModeProtocolHandler.get_response(3, parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>WCGET\x03</DATAS></PACKT>",
        )

    def test_recv_can_handle(self) -> None:
        handler = GeckoGetWatercareModeProtocolHandler()
        self.assertTrue(handler.can_handle(b"GETWC", PARMS))
        self.assertTrue(handler.can_handle(b"WCGET", PARMS))
        self.assertFalse(handler.can_handle(b"OTHER", PARMS))

    def test_recv_handle_request(self) -> None:
        handler = GeckoGetWatercareModeProtocolHandler()
        handler.handle(b"GETWC\x01", PARMS)
        self.assertFalse(handler.should_remove_handler)
        self.assertEqual(handler._sequence, 1)

    def test_recv_handle_response(self) -> None:
        handler = GeckoGetWatercareModeProtocolHandler()
        handler.handle(b"WCGET\x03", PARMS)
        self.assertTrue(handler.should_remove_handler)
        self.assertEqual(handler.mode, 3)


class TestGeckoSetWatercareModeHandler(unittest.TestCase):
    """Set watercare mode handler tests."""

    def test_send_construct_request(self) -> None:
        handler = GeckoSetWatercareModeProtocolHandler.set(1, 2, parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>SETWC\x01\x02</DATAS></PACKT>",
        )

    def test_send_construct_response(self) -> None:
        handler = GeckoSetWatercareModeProtocolHandler.set_response(2, parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>WCSET\x02</DATAS></PACKT>",
        )

    def test_recv_can_handle(self) -> None:
        handler = GeckoSetWatercareModeProtocolHandler()
        self.assertTrue(handler.can_handle(b"SETWC", PARMS))
        self.assertTrue(handler.can_handle(b"WCSET", PARMS))
        self.assertFalse(handler.can_handle(b"OTHER", PARMS))

    def test_recv_handle_request(self) -> None:
        handler = GeckoSetWatercareModeProtocolHandler()
        handler.handle(b"SETWC\x01\x02", PARMS)
        self.assertFalse(handler.should_remove_handler)
        self.assertEqual(handler._sequence, 1)
        self.assertEqual(handler.mode, 2)

    def test_recv_handle_response(self) -> None:
        handler = GeckoSetWatercareModeProtocolHandler()
        handler.handle(b"WCSET\x02", PARMS)
        self.assertTrue(handler.should_remove_handler)
        self.assertEqual(handler.mode, 2)


class TestGeckoGetWatercareScheduleListHandler(unittest.TestCase):
    """Get Watercare schedule list tests."""

    def test_send_construct_request(self) -> None:
        handler = GeckoGetWatercareScheduleListProtocolHandler.get(1, parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>REQWC\x01</DATAS></PACKT>",
        )

    def test_send_construct_response(self) -> None:
        handler = GeckoGetWatercareScheduleListProtocolHandler.get_response(
            GeckoWatercareScheduleManager(), parms=PARMS
        )
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>WCREQ\x00\x00\x00\x01\x00\x00\x06\x00\x00\x00\x00\x02\x01\x00"
            b"\x01\x05"
            b"\x06\x00\x12\x00\x03\x01\x00\x00\x06\x06\x00\x12\x00\x04\x01\x00"
            b"\x01\x05\x00\x00\x00\x00</DATAS></PACKT>",
        )

    def test_recv_can_handle(self) -> None:
        handler = GeckoGetWatercareScheduleListProtocolHandler()
        self.assertTrue(handler.can_handle(b"REQWC", PARMS))
        self.assertTrue(handler.can_handle(b"WCREQ", PARMS))
        self.assertFalse(handler.can_handle(b"OTHER", PARMS))

    def test_recv_handle_request(self) -> None:
        handler = GeckoGetWatercareScheduleListProtocolHandler()
        handler.handle(b"REQWC\x01", PARMS)
        self.assertFalse(handler.should_remove_handler)
        self.assertEqual(handler._sequence, 1)

    def test_recv_handle_response(self) -> None:
        handler = GeckoGetWatercareScheduleListProtocolHandler()
        handler.handle(b"WCREQ\x02", PARMS)
        self.assertTrue(handler.should_remove_handler)


if __name__ == "__main__":
    unittest.main()



"""Main module. Supports python -m geckolib syntax."""

# ruff: noqa: T201

import logging
from pathlib import Path
from sys import argv

from geckolib import CUI, GeckoShell, GeckoSimulator


def usage() -> None:
    """Print usage."""
    print("Usage: python3 -m geckolib <shell|simulator|cui> [client args]")


def install_logging(command: str) -> None:
    """Everyone needs logging, you say when, you say where, you say how much."""
    Path(f"{command}.log").unlink(True)  # noqa: FBT003
    file_logger = logging.FileHandler(f"{command}.log")
    file_logger.setLevel(logging.DEBUG)
    file_logger.setFormatter(
        logging.Formatter("%(asctime)s %(name)s %(levelname)s %(message)s")
    )
    logging.getLogger().addHandler(file_logger)
    logging.getLogger().setLevel(logging.DEBUG)


if len(argv) == 1:
    usage()
elif argv[1] == "shell":
    install_logging(argv[1])
    GeckoShell.run(argv[2:])
elif argv[1] == "simulator":
    install_logging(argv[1])
    GeckoSimulator.run(["load ../tests/snapshots/default.snapshot", "start"] + argv[2:])
elif argv[1] == "cui":
    install_logging(argv[1])
    CUI.launch()
else:
    usage()



"""Unit tests for the UDP protocol handlers."""  # noqa: INP001

# ruff: noqa: E501

import unittest
import unittest.mock

from context import (
    GeckoPingProtocolHandler,
)

PARMS = (1, 2, b"SRCID", b"DESTID")


class TestGeckoPingHandler(unittest.TestCase):
    """Ping handler tests."""

    def test_send_construct_request(self) -> None:
        handler = GeckoPingProtocolHandler.request(parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>APING</DATAS></PACKT>",
        )

    def test_send_construct_response(self) -> None:
        handler = GeckoPingProtocolHandler.response(parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>APING\x00</DATAS></PACKT>",
        )

    def test_recv_can_handle(self) -> None:
        handler = GeckoPingProtocolHandler.request(parms=PARMS)
        self.assertTrue(handler.can_handle(b"APING", PARMS))
        self.assertIsNone(handler._sequence)

    def test_recv_handle(self) -> None:
        handler = GeckoPingProtocolHandler.request(parms=PARMS)
        handler.handle(b"APING\x00", PARMS)
        self.assertFalse(handler.should_remove_handler)
        self.assertEqual(handler._sequence, 0)


if __name__ == "__main__":
    unittest.main()



"""GeckoConfigStruct - A class to manage the ConfigStruct for 'InMixExtended v3'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
)


class GeckoConfigStruct:
    """Config Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the config struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the config struct class version."""
        return 3

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return []

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "InMixExtended-Zone1Led": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMixExtended-Zone1Led",
                696,
                0,
                ["RGB", "WHITE"],
                None,
                2,
                None,
            ),
            "InMixExtended-Zone1Type": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMixExtended-Zone1Type",
                696,
                1,
                ["NORMAL", "STATUS"],
                None,
                2,
                None,
            ),
            "InMixExtended-Zone1Supply": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMixExtended-Zone1Supply",
                696,
                2,
                ["12V", "5V"],
                None,
                2,
                None,
            ),
            "InMixExtended-Zone1UseOut1": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMixExtended-Zone1UseOut1", 696, 4, None
            ),
            "InMixExtended-Zone1UseOut2": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMixExtended-Zone1UseOut2", 696, 5, None
            ),
            "InMixExtended-Zone1UseOut3": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMixExtended-Zone1UseOut3", 696, 6, None
            ),
            "InMixExtended-Zone1UseOut4": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMixExtended-Zone1UseOut4", 696, 7, None
            ),
            "InMixExtended-Zone2Led": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMixExtended-Zone2Led",
                697,
                0,
                ["RGB", "WHITE"],
                None,
                2,
                None,
            ),
            "InMixExtended-Zone2Type": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMixExtended-Zone2Type",
                697,
                1,
                ["NORMAL", "STATUS"],
                None,
                2,
                None,
            ),
            "InMixExtended-Zone2Supply": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMixExtended-Zone2Supply",
                697,
                2,
                ["12V", "5V"],
                None,
                2,
                None,
            ),
            "InMixExtended-Zone2UseOut1": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMixExtended-Zone2UseOut1", 697, 4, None
            ),
            "InMixExtended-Zone2UseOut2": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMixExtended-Zone2UseOut2", 697, 5, None
            ),
            "InMixExtended-Zone2UseOut3": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMixExtended-Zone2UseOut3", 697, 6, None
            ),
            "InMixExtended-Zone2UseOut4": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMixExtended-Zone2UseOut4", 697, 7, None
            ),
            "InMixExtended-Zone3Led": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMixExtended-Zone3Led",
                698,
                0,
                ["RGB", "WHITE"],
                None,
                2,
                None,
            ),
            "InMixExtended-Zone3Type": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMixExtended-Zone3Type",
                698,
                1,
                ["NORMAL", "STATUS"],
                None,
                2,
                None,
            ),
            "InMixExtended-Zone3Supply": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMixExtended-Zone3Supply",
                698,
                2,
                ["12V", "5V"],
                None,
                2,
                None,
            ),
            "InMixExtended-Zone3UseOut1": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMixExtended-Zone3UseOut1", 698, 4, None
            ),
            "InMixExtended-Zone3UseOut2": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMixExtended-Zone3UseOut2", 698, 5, None
            ),
            "InMixExtended-Zone3UseOut3": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMixExtended-Zone3UseOut3", 698, 6, None
            ),
            "InMixExtended-Zone3UseOut4": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMixExtended-Zone3UseOut4", 698, 7, None
            ),
            "InMixExtended-Zone4Led": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMixExtended-Zone4Led",
                699,
                0,
                ["RGB", "WHITE"],
                None,
                2,
                None,
            ),
            "InMixExtended-Zone4Type": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMixExtended-Zone4Type",
                699,
                1,
                ["NORMAL", "STATUS"],
                None,
                2,
                None,
            ),
            "InMixExtended-Zone4Supply": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMixExtended-Zone4Supply",
                699,
                2,
                ["12V", "5V"],
                None,
                2,
                None,
            ),
            "InMixExtended-Zone4UseOut1": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMixExtended-Zone4UseOut1", 699, 4, None
            ),
            "InMixExtended-Zone4UseOut2": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMixExtended-Zone4UseOut2", 699, 5, None
            ),
            "InMixExtended-Zone4UseOut3": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMixExtended-Zone4UseOut3", 699, 6, None
            ),
            "InMixExtended-Zone4UseOut4": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMixExtended-Zone4UseOut4", 699, 7, None
            ),
            "InMixExtended-NumberOfZones": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMixExtended-NumberOfZones",
                700,
                0,
                ["0", "1", "2", "3", "4"],
                None,
                5,
                "ALL",
            ),
            "InMixExtended-StatusLightNormalColor": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMixExtended-StatusLightNormalColor",
                700,
                3,
                ["BLUE", "WHITE", "GREEN", "PURPLE", "YELLOW", "CYAN", "ORANGE", "OFF"],
                None,
                8,
                "ALL",
            ),
            "InMixExtended-OutputAttribution": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/InMixExtended-OutputAttribution",
                700,
                7,
                None,
            ),
            "InMixExtended-MinimumFadeIntensity": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMixExtended-MinimumFadeIntensity",
                701,
                0,
                [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                ],
                None,
                16,
                "ALL",
            ),
        }



"""Gecko <HELLO> handlers."""

from __future__ import annotations

import logging
import time
from typing import Any

from geckolib.const import GeckoConstants
from geckolib.driver.udp_protocol_handler import GeckoUdpProtocolHandler

HELLO_OPEN = b"<HELLO>"
HELLO_CLOSE = b"</HELLO>"

_LOGGER = logging.getLogger(__name__)


class GeckoHelloProtocolHandler(GeckoUdpProtocolHandler):
    """Gecko <HELLO> handler."""

    def __init__(self, content: bytes, **kwargs: Any) -> None:
        """Initialize the class."""
        super().__init__(**kwargs)
        self._send_bytes = b"".join([HELLO_OPEN, content, HELLO_CLOSE])
        self.was_broadcast_discovery: bool = False
        self._client_identifier: bytes | None = None
        self._spa_identifier: bytes | None = None
        self._spa_name: str | None = None
        self.last_response: float = time.monotonic()

    @staticmethod
    def broadcast_address(static_ip: str | None) -> tuple:
        """Generate a broadcast address."""
        if static_ip is not None:
            return (static_ip, GeckoConstants.INTOUCH2_PORT)
        return (GeckoConstants.BROADCAST_ADDRESS, GeckoConstants.INTOUCH2_PORT)

    @staticmethod
    def broadcast(**kwargs: object) -> GeckoHelloProtocolHandler:
        """Generate a broadcast class."""
        return GeckoHelloProtocolHandler(b"1", **kwargs)

    @staticmethod
    def client(client_identifier: bytes, **kwargs: object) -> GeckoHelloProtocolHandler:
        """Generate a client class."""
        return GeckoHelloProtocolHandler(client_identifier, **kwargs)

    @staticmethod
    def response(
        spa_identifier: bytes, spa_name: str, **kwargs: object
    ) -> GeckoHelloProtocolHandler:
        """Generate a server response class."""
        return GeckoHelloProtocolHandler(
            b"".join(
                [spa_identifier, b"|", spa_name.encode(GeckoConstants.MESSAGE_ENCODING)]
            ),
            **kwargs,
        )

    @property
    def client_identifier(self) -> bytes:
        """Get the client identifier or raise exception if not set."""
        assert self._client_identifier is not None  # noqa: S101
        return self._client_identifier

    @property
    def spa_identifier(self) -> bytes:
        """Get the spa identifier or raise exception if not set."""
        assert self._spa_identifier is not None  # noqa: S101
        return self._spa_identifier

    @property
    def spa_name(self) -> str:
        """Get the spa name or raise exception if not set."""
        assert self._spa_name is not None  # noqa: S101
        return self._spa_name

    def can_handle(self, received_bytes: bytes, _sender: tuple) -> bool:
        """Determine if we can handle the packet bytes."""
        return received_bytes.startswith(HELLO_OPEN) and received_bytes.endswith(
            HELLO_CLOSE
        )

    def handle(self, received_bytes: bytes, _sender: tuple) -> None:
        """Handle the packet bytes."""
        content = received_bytes[7:-8]
        self.was_broadcast_discovery = False
        self._client_identifier = self._spa_identifier = self._spa_name = None

        if content == b"1":
            self.was_broadcast_discovery = True

        elif content.startswith((b"IOS", b"AND")):
            self._client_identifier = content

        elif not content.find(b"|"):
            self._spa_identifier = content
            self._spa_name = "Unnamed SPA"

        else:
            self._spa_identifier, spa_name = content.split(b"|")
            self._spa_name = spa_name.decode(GeckoConstants.MESSAGE_ENCODING)

    def __repr__(self) -> str:
        """Get string representation of the class."""
        return (
            f"{super().__repr__()} (was_broadcast={self.was_broadcast_discovery},"
            f" client_id={self._client_identifier!r}, spa_id={self._spa_identifier!r},"
            f" spa_name={self._spa_name} )"
        )



"""Gecko driver."""

from .accessor import (
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoTempStructAccessor,
    GeckoTimeStructAccessor,
    GeckoWordStructAccessor,
)
from .async_spastruct import GeckoAsyncStructure
from .async_udp_protocol import GeckoAsyncUdpProtocol
from .observable import Observable
from .protocol import (
    GeckoAsyncPartialStatusBlockProtocolHandler,
    GeckoConfigFileProtocolHandler,
    GeckoGetChannelProtocolHandler,
    GeckoGetWatercareModeProtocolHandler,
    GeckoGetWatercareScheduleListProtocolHandler,
    GeckoHelloProtocolHandler,
    GeckoPackCommandProtocolHandler,
    GeckoPacketProtocolHandler,
    GeckoPingProtocolHandler,
    GeckoRemindersProtocolHandler,
    GeckoReminderType,
    GeckoRFErrProtocolHandler,
    GeckoSetWatercareModeProtocolHandler,
    GeckoStatusBlockProtocolHandler,
    GeckoUnhandledProtocolHandler,
    GeckoUpdateFirmwareProtocolHandler,
    GeckoVersionProtocolHandler,
    GeckoWatercareErrorHandler,
    GeckoWatercareScheduleManager,
)
from .udp_protocol_handler import GeckoUdpProtocolHandler

__all__ = [
    "GeckoAsyncPartialStatusBlockProtocolHandler",
    "GeckoAsyncStructure",
    "GeckoAsyncUdpProtocol",
    "GeckoBoolStructAccessor",
    "GeckoByteStructAccessor",
    "GeckoConfigFileProtocolHandler",
    "GeckoEnumStructAccessor",
    "GeckoGetChannelProtocolHandler",
    "GeckoGetWatercareModeProtocolHandler",
    "GeckoGetWatercareScheduleListProtocolHandler",
    "GeckoHelloProtocolHandler",
    "GeckoPackCommandProtocolHandler",
    "GeckoPacketProtocolHandler",
    "GeckoPingProtocolHandler",
    "GeckoRFErrProtocolHandler",
    "GeckoReminderType",
    "GeckoRemindersProtocolHandler",
    "GeckoSetWatercareModeProtocolHandler",
    "GeckoStatusBlockProtocolHandler",
    "GeckoStructAccessor",
    "GeckoTempStructAccessor",
    "GeckoTimeStructAccessor",
    "GeckoUdpProtocolHandler",
    "GeckoUnhandledProtocolHandler",
    "GeckoUpdateFirmwareProtocolHandler",
    "GeckoVersionProtocolHandler",
    "GeckoWatercareErrorHandler",
    "GeckoWatercareScheduleManager",
    "GeckoWordStructAccessor",
    "Observable",
]



"""Unit tests for the SpaMan class."""  # noqa: INP001

from typing import Any
from unittest import IsolatedAsyncioTestCase, main
from unittest.mock import patch

from context import GeckoAsyncSpaDescriptor, GeckoAsyncSpaMan, GeckoSpaEvent


class SpaManImpl(GeckoAsyncSpaMan):
    """Spa Manager to test with."""

    def __init__(self) -> None:
        """Initialize the spaman class."""
        super().__init__("CLIENT_UUID", spa_identifier="TestID")
        self.events = []

    async def handle_event(self, event: GeckoSpaEvent, **_kwargs: object) -> None:
        self.events.append(event)


mock_spa_descriptor = GeckoAsyncSpaDescriptor(b"TestID", "Test Name", (1, 2))
mock_spas = [mock_spa_descriptor]


async def mock_discover(self: Any) -> None:
    """Mock discovery."""
    await self._event_handler(GeckoSpaEvent.LOCATING_DISCOVERED_SPA)


async def mock_connect(self: Any) -> None:
    """Mock connection."""
    # await self._event_handler(GeckoSpaEvent.CONNECTION_SPA_COMPLETE)  # noqa: ERA001


@patch("context.GeckoAsyncSpa._connect", mock_connect)
@patch("context.GeckoAsyncLocator.discover", mock_discover)
@patch("context.GeckoAsyncLocator.spas", mock_spas)
class TestSpaMan(IsolatedAsyncioTestCase):
    """Test the SpaMan class."""

    def setUp(self) -> None:
        self.spaman = SpaManImpl()

    async def asyncSetUp(self) -> None:
        await self.spaman.__aenter__()

    async def asyncTearDown(self) -> None:
        await self.spaman.__aexit__(None)

    def tearDown(self) -> None:
        del self.spaman

    #####################################################

    def test_facade_on_start(self) -> None:
        self.assertIsNone(self.spaman.facade)

    async def test_locate_spas(self) -> None:
        spas = await self.spaman.async_locate_spas()
        self.assertEqual(len(spas), 1)
        self.assertEqual(spas[0].identifier_as_string, "TestID")
        self.assertEqual(spas[0].name, "Test Name")
        self.assertListEqual(
            self.spaman.events,
            [
                GeckoSpaEvent.SPA_MAN_ENTER,
                GeckoSpaEvent.LOCATING_STARTED,
                GeckoSpaEvent.LOCATING_STARTED,
                GeckoSpaEvent.LOCATING_DISCOVERED_SPA,
                GeckoSpaEvent.LOCATING_FINISHED,
            ],
        )

    async def test_connect_spa(self) -> None:
        facade = await self.spaman.async_connect_to_spa(mock_spa_descriptor)
        self.assertListEqual(
            self.spaman.events,
            [
                GeckoSpaEvent.SPA_MAN_ENTER,
                GeckoSpaEvent.LOCATING_STARTED,
                GeckoSpaEvent.CLIENT_HAS_STATUS_SENSOR,
                GeckoSpaEvent.CLIENT_HAS_RECONNECT_BUTTON,
                GeckoSpaEvent.CONNECTION_STARTED,
                GeckoSpaEvent.CONNECTION_FINISHED,
            ],
        )
        self.assertIsNone(facade)


if __name__ == "__main__":
    main()



"""Gecko handler for unhanded verbs."""

import logging
from typing import Any

from geckolib.driver.async_udp_protocol import GeckoAsyncUdpProtocol
from geckolib.driver.udp_protocol_handler import GeckoUdpProtocolHandler

_LOGGER = logging.getLogger(__name__)


class GeckoUnhandledProtocolHandler(GeckoUdpProtocolHandler):
    """Protocol handler to deal with unhandled or unsolicited messages."""

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the default handler."""
        super().__init__(**kwargs)

    def can_handle(self, _received_bytes: bytes, _sender: tuple) -> None:
        """Will never get called."""

    def handle(self, received_bytes: bytes, sender: tuple) -> None:
        """Go on then, handle it."""

    def _can_handle(
        self, protocol: GeckoAsyncUdpProtocol, _received_bytes: bytes, _sender: tuple
    ) -> bool:
        # If the protocol queue is marked, then we can handle it because no
        # one else can
        if protocol.queue.is_marked:
            _LOGGER.debug(
                "Packet %s from %s is unhandled and will be ignored",
                _received_bytes,
                _sender,
            )
            return True
        # Mark the queue and let others have a go
        protocol.queue.mark()
        return False



"""Automation lockmode class."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from geckolib.const import GeckoConstants

from .select import GeckoSelect

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade

_LOGGER = logging.getLogger(__name__)


class GeckoLockMode(GeckoSelect):
    """A select object can select between options and can report the current state."""

    def __init__(self, facade: GeckoAsyncFacade) -> None:
        """Initialize the heatpump class."""
        super().__init__(facade, "Lock Mode", GeckoConstants.KEY_LOCKMODE)
        # Set of mappings of constants to UI options
        self.mapping = {
            "UNLOCK": "Unlocked",
            "PARTIAL": "Partial Lock",
            "FULL": "Full Lock",
        }
        self.reverse = {v: k for k, v in self.mapping.items()}

    @property
    def state(self) -> str:
        """Get the current state via the mapping."""
        return self.mapping[self._accessor.value]

    async def async_set_state(self, new_state: str) -> None:
        """Set the state of the select entity."""
        if new_state in self.reverse:
            new_state = self.reverse[new_state]
        await self._accessor.async_set_value(new_state)

    @property
    def states(self) -> list[str]:
        """Get the possible states."""
        return list(self.mapping.values())



"""Unit tests for the UDP protocol handlers."""  # noqa: INP001

# ruff: noqa: E501

import unittest
import unittest.mock

from context import (
    GeckoPacketProtocolHandler,
)

PARMS = (1, 2, b"SRCID", b"DESTID")


class TestGeckoPacketHandler(unittest.TestCase):
    """Packet handler tests."""

    def setUp(self) -> None:
        """Set up the class."""
        self.handler = GeckoPacketProtocolHandler(content=b"CONTENT", parms=PARMS)

    def test_recv_can_handle(self) -> None:
        self.assertTrue(self.handler.can_handle(b"<PACKT></PACKT>", ()))
        self.assertFalse(self.handler.can_handle(b"<PACKT></PACKT", ()))
        self.assertFalse(self.handler.can_handle(b"<PACKT></PACKT> ", ()))
        self.assertFalse(self.handler.can_handle(b"<SOMETHING>", ()))

    def test_recv_extract_ok(self) -> None:
        self.assertFalse(
            self.handler.handle(
                b"<PACKT><SRCCN>SRCID</SRCCN><DESCN>DESTID</DESCN>"
                b"<DATAS>DATA</DATAS></PACKT>",
                (1, 2),
            )
        )
        assert self.handler.parms is not None
        self.assertTupleEqual(self.handler.parms, PARMS)
        self.assertEqual(self.handler.packet_content, b"DATA")

    def test_send_construct(self) -> None:
        self.assertEqual(
            self.handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>CONTENT</DATAS></PACKT>",
        )
        assert self.handler.parms is not None
        self.assertEqual(self.handler.parms[0], 1)
        self.assertEqual(self.handler.parms[1], 2)


if __name__ == "__main__":
    unittest.main()



"""Spa state."""

from __future__ import annotations

from enum import Enum


class GeckoSpaState(Enum):
    """Spa, locator or spa manager state."""

    IDLE = 1
    """Idle state is when the spa manager is first initialized"""

    LOCATING_SPAS = 2
    """State when the spa manager is locating spas on the network"""

    LOCATED_SPAS = 3
    """State when the spa manager has finished locating spas"""

    CONNECTING = 4
    """State when the spa is going through connection protocol"""

    SPA_READY = 5
    """State when the spa is ready to have a facade built"""

    CONNECTED = 10
    """State when the spa manager is connected to a spa successfully"""

    ERROR_SPA_NOT_FOUND = 50
    """State when the spa cannot be found on the the network"""
    ERROR_NEEDS_ATTENTION = 51
    """State when the user needs to attend"""
    ERROR_PING_MISSED = 52
    """State when a ping was missed, can auto-connect on restore"""
    ERROR_RF_FAULT = 53
    """State when the EN module reports it can't talk to the CO module"""

    ERROR_NOT_SUPPORTED = 100
    """State when the pack isn't supported, but it looks like it talks the talk"""

    @staticmethod
    def to_string(state: GeckoSpaState) -> str:  # noqa: PLR0911
        """Convert GeckoSpaState to string."""
        if state == GeckoSpaState.CONNECTED:
            return "Connected"
        if state == GeckoSpaState.CONNECTING:
            return "Connecting..."
        if state == GeckoSpaState.ERROR_RF_FAULT:
            return "Lost contact with spa (RFERR)"
        if state == GeckoSpaState.ERROR_PING_MISSED:
            return "Lost contact with in.touch2 module"
        if state == GeckoSpaState.ERROR_NEEDS_ATTENTION:
            return "Needs attention, check logs"
        if state == GeckoSpaState.LOCATING_SPAS:
            return "Searching for spas..."
        if state == GeckoSpaState.LOCATED_SPAS:
            return "Choose spa"
        if state == GeckoSpaState.ERROR_SPA_NOT_FOUND:
            return "Cannot find spa, check logs"
        if state == GeckoSpaState.ERROR_NOT_SUPPORTED:
            return "Device not supported, check logs"
        return f"{state}"



"""Gecko UPDTS/SUPDT handlers."""

from __future__ import annotations

import logging
import struct
from typing import Any

from geckolib.config import GeckoConfig

from .packet import GeckoPacketProtocolHandler

UPDTS_VERB = b"UPDTS"
SUPDT_VERB = b"SUPDT"

_LOGGER = logging.getLogger(__name__)


class GeckoUpdateFirmwareProtocolHandler(GeckoPacketProtocolHandler):
    """Handle UPDTS/SUPDT."""

    @staticmethod
    def request(seq: int, **kwargs: Any) -> GeckoUpdateFirmwareProtocolHandler:
        """Generate the request."""
        return GeckoUpdateFirmwareProtocolHandler(
            content=b"".join([UPDTS_VERB, struct.pack(">B", seq)]),
            timeout=GeckoConfig.PROTOCOL_TIMEOUT_IN_SECONDS,
            on_retry_failed=GeckoPacketProtocolHandler.default_retry_failed_handler,
            **kwargs,
        )

    @staticmethod
    def response(**kwargs: Any) -> GeckoUpdateFirmwareProtocolHandler:
        """Generate the response."""
        return GeckoUpdateFirmwareProtocolHandler(
            content=b"".join(
                [
                    SUPDT_VERB,
                    b"\x00",
                ]
            ),
            **kwargs,
        )

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the class."""
        super().__init__(**kwargs)

    def can_handle(self, received_bytes: bytes, _sender: tuple) -> bool:
        """Is this verb handles."""
        return received_bytes.startswith((UPDTS_VERB, SUPDT_VERB))

    def handle(self, received_bytes: bytes, _sender: tuple) -> None:
        """Handle the verb."""
        remainder = received_bytes[5:]
        if received_bytes.startswith(UPDTS_VERB):
            self._sequence = struct.unpack(">B", remainder[0:1])[0]
            return  # Stay in the handler list
        # Otherwise must be SUPDT
        self._should_remove_handler = True
